# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2020 Canonical Ltd
# Copyright (C) 2016 Harald Sitter <sitter@kde.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pathlib
import stat
from unittest import mock

import pytest
from testtools.matchers import Equals, HasLength

import snapcraft
from snapcraft.internal import errors, meta
from snapcraft.plugins.v1 import autotools, make

from . import PluginsV1BaseTestCase


@pytest.fixture
def options():
    class Options:
        configflags = []
        install_via = "destdir"
        disable_parallel = False
        makefile = None
        make_parameters = []
        make_install_var = "DESTDIR"
        artifacts = []

    return Options()


class AutotoolsPluginTestCase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class Options:
            configflags = []
            install_via = "destdir"
            disable_parallel = False
            makefile = None
            make_parameters = []
            make_install_var = "DESTDIR"
            artifacts = []

        self.options = Options()

    def test_schema(self):
        schema = autotools.AutotoolsPlugin.schema()

        # Verify the presence of all properties
        properties = schema["properties"]
        self.assertTrue(
            "configflags" in properties,
            'Expected "configflags" to be included in properties',
        )
        self.assertTrue(
            "install-via" in properties,
            'Expected "install-via" to be included in properties',
        )

        # Check configflags property
        configflags = properties["configflags"]
        for item in ["type", "minitems", "items", "default"]:
            self.assertTrue(
                item in configflags,
                'Expected "{}" to be included in "configflags"'.format(item),
            )

        configflags_type = configflags["type"]
        self.assertThat(
            configflags_type,
            Equals("array"),
            'Expected "configflags" "type" to be "array", but it '
            'was "{}"'.format(configflags_type),
        )

        configflags_minitems = configflags["minitems"]
        self.assertThat(
            configflags_minitems,
            Equals(1),
            'Expected "configflags" "minitems" to be 1, but '
            "it was {}".format(configflags_minitems),
        )

        configflags_default = configflags["default"]
        self.assertThat(
            configflags_default,
            Equals([]),
            'Expected "configflags" "default" to be [], but '
            "it was {}".format(configflags_default),
        )

        configflags_items = configflags["items"]
        self.assertTrue(
            "type" in configflags_items,
            'Expected "type" to be included in "configflags" ' '"items"',
        )

        configflags_items_type = configflags_items["type"]
        self.assertThat(
            configflags_items_type,
            Equals("string"),
            'Expected "configflags" "items" "type" to be '
            '"string", but it was "{}"'.format(configflags_items_type),
        )

        # Check install-via property
        installvia = properties["install-via"]
        self.assertTrue(
            "enum" in installvia, 'Expected "enum" to be included in "install-via"'
        )
        self.assertTrue(
            "default" in installvia,
            'Expected "default" to be included in "install-via"',
        )

        installvia_enum = installvia["enum"]
        # Using sets for order independence in the comparison
        self.assertThat(set(installvia_enum), Equals(set(["destdir", "prefix"])))

        installvia_default = installvia["default"]
        self.assertThat(
            installvia_default,
            Equals("destdir"),
            'Expected "install-via" "default" to be "destdir", '
            'but it was "{}"'.format(installvia_default),
        )

    def test_get_build_properties(self):
        expected_build_properties = make.MakePlugin.get_build_properties() + (
            ["configflags", "install-via"]
        )
        resulting_build_properties = autotools.AutotoolsPlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)

    def test_install_via_invalid_enum(self):
        self.options.install_via = "invalid"
        raised = self.assertRaises(
            RuntimeError,
            autotools.AutotoolsPlugin,
            "test-part",
            self.options,
            self.project,
        )

        self.assertThat(
            str(raised), Equals('Unsupported installation method: "invalid"')
        )

    def build_with_configure(self):
        plugin = autotools.AutotoolsPlugin("test-part", self.options, self.project)
        os.makedirs(plugin.builddir)

        # Create both configure and autogen.sh.
        # Configure should take precedence.
        open(os.path.join(plugin.builddir, "configure"), "w").close()
        open(os.path.join(plugin.builddir, "autogen.sh"), "w").close()

        plugin.build()

        return plugin

    @mock.patch.object(autotools.AutotoolsPlugin, "run")
    def test_build_configure_with_destdir(self, run_mock):
        plugin = self.build_with_configure()

        self.assertThat(run_mock.call_count, Equals(3))
        run_mock.assert_has_calls(
            [
                mock.call(["./configure", "--prefix="]),
                mock.call(["make", "-j2"], env=None),
                mock.call(
                    ["make", "install", "DESTDIR={}".format(plugin.installdir)],
                    env=None,
                ),
            ]
        )

    @mock.patch.object(autotools.AutotoolsPlugin, "run")
    def test_build_configure_with_prefix(self, run_mock):
        self.options.install_via = "prefix"
        plugin = self.build_with_configure()

        self.assertThat(run_mock.call_count, Equals(3))
        run_mock.assert_has_calls(
            [
                mock.call(["./configure", "--prefix={}".format(plugin.installdir)]),
                mock.call(["make", "-j2"], env=None),
                mock.call(["make", "install"], env=None),
            ]
        )

    def build_with_autogen(self, files=None, dirs=None):
        plugin = autotools.AutotoolsPlugin("test-part", self.options, self.project)
        os.makedirs(plugin.builddir)

        if not files:
            files = ["autogen.sh"]
        if not dirs:
            dirs = []

        # No configure-- only autogen.sh. Make sure it's executable.
        for filename in files:
            open(os.path.join(plugin.builddir, filename), "w").close()
            os.chmod(
                os.path.join(plugin.builddir, filename),
                stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR,
            )
        for directory in dirs:
            os.makedirs(os.path.join(plugin.builddir, directory))

        plugin.build()

        return plugin

    @mock.patch.object(autotools.AutotoolsPlugin, "run")
    def test_build_autogen_with_bootstrap_dir(self, run_mock):
        plugin = self.build_with_autogen(files=["README"], dirs=["bootstrap"])

        self.assertThat(run_mock.call_count, Equals(4))
        run_mock.assert_has_calls(
            [
                mock.call(["autoreconf", "-i"]),
                mock.call(["./configure", "--prefix="]),
                mock.call(["make", "-j2"], env=None),
                mock.call(
                    ["make", "install", "DESTDIR={}".format(plugin.installdir)],
                    env=None,
                ),
            ]
        )

    @mock.patch.object(autotools.AutotoolsPlugin, "run")
    def test_build_autogen_with_destdir(self, run_mock):
        plugin = self.build_with_autogen()

        self.assertThat(run_mock.call_count, Equals(4))
        run_mock.assert_has_calls(
            [
                mock.call(["env", "NOCONFIGURE=1", "./autogen.sh"]),
                mock.call(["./configure", "--prefix="]),
                mock.call(["make", "-j2"], env=None),
                mock.call(
                    ["make", "install", "DESTDIR={}".format(plugin.installdir)],
                    env=None,
                ),
            ]
        )

    @mock.patch.object(autotools.AutotoolsPlugin, "run")
    def test_build_bootstrap_with_destdir(self, run_mock):
        plugin = self.build_with_autogen(files=["bootstrap"])

        self.assertThat(run_mock.call_count, Equals(4))
        run_mock.assert_has_calls(
            [
                mock.call(["env", "NOCONFIGURE=1", "./bootstrap"]),
                mock.call(["./configure", "--prefix="]),
                mock.call(["make", "-j2"], env=None),
                mock.call(
                    ["make", "install", "DESTDIR={}".format(plugin.installdir)],
                    env=None,
                ),
            ]
        )

    @mock.patch.object(autotools.AutotoolsPlugin, "run")
    def test_build_bootstrap_and_autogen_with_destdir(self, run_mock):
        plugin = self.build_with_autogen(files=["bootstrap", "autogen.sh"])

        self.assertThat(run_mock.call_count, Equals(4))
        run_mock.assert_has_calls(
            [
                mock.call(["env", "NOCONFIGURE=1", "./autogen.sh"]),
                mock.call(["./configure", "--prefix="]),
                mock.call(["make", "-j2"], env=None),
                mock.call(
                    ["make", "install", "DESTDIR={}".format(plugin.installdir)],
                    env=None,
                ),
            ]
        )

    @mock.patch.object(autotools.AutotoolsPlugin, "run")
    def test_build_autogen_with_prefix(self, run_mock):
        self.options.install_via = "prefix"
        plugin = self.build_with_autogen()

        self.assertThat(run_mock.call_count, Equals(4))
        run_mock.assert_has_calls(
            [
                mock.call(["env", "NOCONFIGURE=1", "./autogen.sh"]),
                mock.call(["./configure", "--prefix={}".format(plugin.installdir)]),
                mock.call(["make", "-j2"], env=None),
                mock.call(["make", "install"], env=None),
            ]
        )

    def build_with_autoreconf(self):
        plugin = autotools.AutotoolsPlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)

        # No configure or autogen.sh.

        plugin.build()

        return plugin

    @mock.patch.object(autotools.AutotoolsPlugin, "run")
    def test_build_autoreconf_with_destdir(self, run_mock):
        plugin = self.build_with_autoreconf()

        self.assertThat(run_mock.call_count, Equals(4))
        run_mock.assert_has_calls(
            [
                mock.call(["autoreconf", "-i"]),
                mock.call(["./configure", "--prefix="]),
                mock.call(["make", "-j2"], env=None),
                mock.call(
                    ["make", "install", "DESTDIR={}".format(plugin.installdir)],
                    env=None,
                ),
            ]
        )

    @mock.patch.object(autotools.AutotoolsPlugin, "run")
    def test_build_autoreconf_with_prefix(self, run_mock):
        self.options.install_via = "prefix"
        plugin = self.build_with_autoreconf()

        self.assertThat(run_mock.call_count, Equals(4))
        run_mock.assert_has_calls(
            [
                mock.call(["autoreconf", "-i"]),
                mock.call(["./configure", "--prefix={}".format(plugin.installdir)]),
                mock.call(["make", "-j2"], env=None),
                mock.call(["make", "install"], env=None),
            ]
        )

    @mock.patch.object(autotools.AutotoolsPlugin, "run")
    def test_build_autoreconf_with_disable_parallel(self, run_mock):
        self.options.disable_parallel = True
        plugin = self.build_with_autoreconf()

        self.assertThat(run_mock.call_count, Equals(4))
        run_mock.assert_has_calls(
            [
                mock.call(["autoreconf", "-i"]),
                mock.call(["./configure", "--prefix="]),
                mock.call(["make", "-j1"], env=None),
                mock.call(
                    ["make", "install", "DESTDIR={}".format(plugin.installdir)],
                    env=None,
                ),
            ]
        )

    @mock.patch("sys.stdout")
    def test_build_nonexecutable_autogen(self, stdout_mock):
        plugin = autotools.AutotoolsPlugin("test-part", self.options, self.project)
        os.makedirs(plugin.sourcedir)

        # Make a non-executable autogen.sh
        with open(os.path.join(plugin.sourcedir, "autogen.sh"), "w") as f:
            f.write("#!/bin/sh")

        patcher = mock.patch.object(autotools.AutotoolsPlugin, "run")
        run_mock = patcher.start()

        # We want to mock out every run() call except the one to autogen
        def _run(cmd, env=None):
            if "./autogen.sh" in cmd:
                patcher.stop()
                output = plugin.run(cmd, env=env)
                patcher.start()
                return output

        run_mock.side_effect = _run

        # An exception will be raised if build can't handle the non-executable
        # autogen.
        plugin.build()

    def test_fileset_ignores(self):
        plugin = autotools.AutotoolsPlugin("test-part", self.options, self.project)
        expected_fileset = ["-**/*.la"]
        fileset = plugin.snap_fileset()
        self.assertListEqual(expected_fileset, fileset)

    def test_unsupported_base(self):
        self.project._snap_meta.base = "unsupported-base"

        raised = self.assertRaises(
            errors.PluginBaseError,
            autotools.AutotoolsPlugin,
            "test-part",
            self.options,
            self.project,
        )

        self.assertThat(raised.part_name, Equals("test-part"))
        self.assertThat(raised.base, Equals("unsupported-base"))


@pytest.mark.parametrize(
    "deb_arch,triplet",
    [
        ("armhf", "arm-linux-gnueabihf"),
        ("arm64", "aarch64-linux-gnu"),
        ("i386", "i386-linux-gnu"),
        ("amd64", "x86_64-linux-gnu"),
        ("ppc64el", "powerpc64le-linux-gnu"),
    ],
)
@mock.patch.object(autotools.AutotoolsPlugin, "run")
def test_cross_compile(mock_run, monkeypatch, project, options, deb_arch, triplet):
    monkeypatch.setattr(snapcraft.project.Project, "is_cross_compiling", True)

    project = snapcraft.project.Project(target_deb_arch=deb_arch)
    project._snap_meta = meta.snap.Snap(name="test-snap", base="core18")

    plugin = autotools.AutotoolsPlugin("test-part", options, project)
    plugin.enable_cross_compilation()
    configure_file = pathlib.Path(plugin.builddir) / "configure"
    configure_file.parent.mkdir(parents=True)
    configure_file.touch()

    plugin.build()

    mock_run.assert_has_calls(
        [mock.call(["./configure", "--prefix=", f"--host={triplet}"])]
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from snapcraft.storeapi import errors


class TestSnapcraftException:
    scenarios = (
        (
            "SnapNotFoundError snap_name",
            dict(
                exception_class=errors.SnapNotFoundError,
                kwargs=dict(snap_name="not-found"),
                expected_brief="Snap 'not-found' was not found.",
                expected_resolution="Ensure you have proper access rights for 'not-found'.",
                expected_details=None,
                expected_docs_url=None,
                expected_reportable=False,
            ),
        ),
        (
            "SnapNotFoundError snap_id",
            dict(
                exception_class=errors.SnapNotFoundError,
                kwargs=dict(snap_id="1234567890"),
                expected_brief="Cannot find snap with snap_id '1234567890'.",
                expected_resolution="Ensure you have proper access rights for '1234567890'.",
                expected_details=None,
                expected_docs_url=None,
                expected_reportable=False,
            ),
        ),
        (
            "SnapNotFoundError snap_name, channel and arch",
            dict(
                exception_class=errors.SnapNotFoundError,
                kwargs=dict(
                    snap_name="not-found", channel="default/stable", arch="foo"
                ),
                expected_brief="Snap 'not-found' for architecture 'foo' was not found on channel 'default/stable'.",
                expected_resolution=dedent(
                    """\
                Ensure you have proper access rights for 'not-found'.
                Also ensure the correct channel and architecture was used."""
                ),
                expected_details=None,
                expected_docs_url=None,
                expected_reportable=False,
            ),
        ),
        (
            "SnapNotFoundError snap_name and channel",
            dict(
                exception_class=errors.SnapNotFoundError,
                kwargs=dict(snap_name="not-found", channel="default/stable"),
                expected_brief="Snap 'not-found' was not found on channel 'default/stable'.",
                expected_resolution=dedent(
                    """\
                Ensure you have proper access rights for 'not-found'.
                Also ensure the correct channel was used."""
                ),
                expected_details=None,
                expected_docs_url=None,
                expected_reportable=False,
            ),
        ),
        (
            "SnapNotFoundError snap_name and arch",
            dict(
                exception_class=errors.SnapNotFoundError,
                kwargs=dict(snap_name="not-found", arch="foo"),
                expected_brief="Snap 'not-found' for architecture 'foo' was not found.",
                expected_resolution=dedent(
                    """\
                Ensure you have proper access rights for 'not-found'.
                Also ensure the correct architecture was used."""
                ),
                expected_details=None,
                expected_docs_url=None,
                expected_reportable=False,
            ),
        ),
    )

    def test_snapcraft_exception_handling(
        self,
        exception_class,
        kwargs,
        expected_brief,
        expected_resolution,
        expected_details,
        expected_docs_url,
        expected_reportable,
    ):
        exception = exception_class(**kwargs)
        assert exception.get_brief() == expected_brief
        assert exception.get_resolution() == expected_resolution
        assert exception.get_details() == expected_details
        assert exception.get_docs_url() == expected_docs_url
        assert exception.get_reportable() == expected_reportable



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import hashlib
import json
import logging
import operator
import os
import re
import subprocess
import tempfile
from datetime import datetime
from pathlib import Path
from subprocess import Popen
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, TextIO, Tuple
from urllib.parse import urljoin

from tabulate import tabulate

from snapcraft import storeapi, yaml_utils

# Ideally we would move stuff into more logical components
from snapcraft.cli import echo
from snapcraft.file_utils import (
    calculate_sha3_384,
    get_host_tool_path,
    get_snap_tool_path,
)
from snapcraft.internal import cache, deltas
from snapcraft.internal.deltas.errors import (
    DeltaGenerationError,
    DeltaGenerationTooBigError,
)
from snapcraft.internal.errors import SnapDataExtractionError, ToolMissingError
from snapcraft.storeapi.constants import DEFAULT_SERIES

if TYPE_CHECKING:
    from snapcraft.storeapi._status_tracker import StatusTracker
    from snapcraft.storeapi.v2.channel_map import ChannelMap
    from snapcraft.storeapi.v2.releases import Releases


logger = logging.getLogger(__name__)


def _get_data_from_snap_file(snap_path):
    with tempfile.TemporaryDirectory() as temp_dir:
        unsquashfs_path = get_snap_tool_path("unsquashfs")
        try:
            output = subprocess.check_output(
                [
                    unsquashfs_path,
                    "-d",
                    os.path.join(temp_dir, "squashfs-root"),
                    snap_path,
                    "-e",
                    # cygwin unsquashfs on windows uses unix paths.
                    Path("meta", "snap.yaml").as_posix(),
                ]
            )
        except subprocess.CalledProcessError:
            raise SnapDataExtractionError(os.path.basename(snap_path))
        logger.debug(output)
        with open(
            os.path.join(temp_dir, "squashfs-root", "meta", "snap.yaml")
        ) as yaml_file:
            snap_yaml = yaml_utils.load(yaml_file)
    return snap_yaml


@contextlib.contextmanager
def _get_icon_from_snap_file(snap_path):
    icon_file = None
    with tempfile.TemporaryDirectory() as temp_dir:
        unsquashfs_path = get_snap_tool_path("unsquashfs")
        try:
            output = subprocess.check_output(
                [
                    unsquashfs_path,
                    "-d",
                    os.path.join(temp_dir, "squashfs-root"),
                    snap_path,
                    "-e",
                    "meta/gui",
                ]
            )
        except subprocess.CalledProcessError:
            raise SnapDataExtractionError(os.path.basename(snap_path))
        logger.debug("Output extracting icon from snap: %s", output)
        for extension in ("png", "svg"):
            icon_name = "icon.{}".format(extension)
            icon_path = os.path.join(temp_dir, "squashfs-root", "meta/gui", icon_name)
            if os.path.exists(icon_path):
                icon_file = open(icon_path, "rb")
                break
        try:
            yield icon_file
        finally:
            if icon_file is not None:
                icon_file.close()


def _get_url_from_error(error: storeapi.errors.StoreAccountInformationError) -> str:
    if error.extra:  # type: ignore
        return error.extra[0].get("url")  # type: ignore
    return ""


def _check_dev_agreement_and_namespace_statuses(store) -> None:
    """ Check the agreement and namespace statuses of the dev.
    Fail if either of those conditions is not met.
    Re-raise `StoreAccountInformationError` if we get an error and
    the error is not either of these.
    """
    # Check account information for the `developer agreement` status.
    try:
        store.get_account_information()
    except storeapi.errors.StoreAccountInformationError as e:
        if storeapi.constants.MISSING_AGREEMENT == e.error:  # type: ignore
            # A precaution if store does not return new style error.
            url = _get_url_from_error(e) or urljoin(
                storeapi.constants.STORE_DASHBOARD_URL, "/dev/tos"
            )
            choice = echo.prompt(storeapi.constants.AGREEMENT_INPUT_MSG.format(url))
            if choice in {"y", "Y"}:
                try:
                    store.sign_developer_agreement(latest_tos_accepted=True)
                except:  # noqa LP: #1733003
                    raise storeapi.errors.NeedTermsSignedError(
                        storeapi.constants.AGREEMENT_SIGN_ERROR.format(url)
                    )
            else:
                raise storeapi.errors.NeedTermsSignedError(
                    storeapi.constants.AGREEMENT_ERROR
                )

    # Now check account information for the `namespace` status.
    try:
        store.get_account_information()
    except storeapi.errors.StoreAccountInformationError as e:
        if storeapi.constants.MISSING_NAMESPACE in e.error:  # type: ignore
            # A precaution if store does not return new style error.
            url = _get_url_from_error(e) or urljoin(
                storeapi.constants.STORE_DASHBOARD_URL, "/dev/account"
            )
            raise storeapi.errors.NeedTermsSignedError(
                storeapi.constants.NAMESPACE_ERROR.format(url)
            )
        else:
            raise


def _try_login(
    email: str,
    password: str,
    *,
    store: storeapi.StoreClient,
    save: bool = True,
    packages: Iterable[Dict[str, str]] = None,
    acls: Iterable[str] = None,
    channels: Iterable[str] = None,
    expires: str = None,
    config_fd: TextIO = None,
) -> None:
    try:
        store.login(
            email=email,
            password=password,
            packages=packages,
            acls=acls,
            channels=channels,
            expires=expires,
            config_fd=config_fd,
            save=save,
        )
        if not config_fd:
            print()
            echo.wrapped(storeapi.constants.TWO_FACTOR_WARNING)
    except storeapi.http_clients.errors.StoreTwoFactorAuthenticationRequired:
        one_time_password = echo.prompt("Second-factor auth")
        store.login(
            email=email,
            password=password,
            otp=one_time_password,
            acls=acls,
            packages=packages,
            channels=channels,
            expires=expires,
            config_fd=config_fd,
            save=save,
        )

    # Continue if agreement and namespace conditions are met.
    _check_dev_agreement_and_namespace_statuses(store)


def login(
    *,
    store: storeapi.StoreClient,
    packages: Iterable[Dict[str, str]] = None,
    save: bool = True,
    acls: Iterable[str] = None,
    channels: Iterable[str] = None,
    expires: str = None,
    config_fd: TextIO = None,
) -> bool:
    if not store:
        store = storeapi.StoreClient()

    email = ""
    password = ""

    if not config_fd:
        echo.wrapped("Enter your Ubuntu One e-mail address and password.")
        echo.wrapped(
            "If you do not have an Ubuntu One account, you can create one "
            "at https://snapcraft.io/account"
        )
        email = echo.prompt("Email")
        if os.getenv("SNAPCRAFT_TEST_INPUT"):
            # Integration tests do not work well with hidden input.
            echo.warning("Password will be visible.")
            hide_input = False
        else:
            hide_input = True
        password = echo.prompt("Password", hide_input=hide_input)

    _try_login(
        email,
        password,
        store=store,
        packages=packages,
        acls=acls,
        channels=channels,
        expires=expires,
        config_fd=config_fd,
        save=save,
    )

    return True


def _login_wrapper(method):
    def login_decorator(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except storeapi.http_clients.errors.InvalidCredentialsError:
            print("You are required to login before continuing.")
            login(store=self)
            return method(self, *args, **kwargs)

    return login_decorator


def _register_wrapper(method):
    def register_decorator(self, *args, snap_name: str, **kwargs):
        try:
            return method(self, *args, snap_name=snap_name, **kwargs)
        except storeapi.errors.StoreUploadError as upload_error:
            if "resource-not-found" not in upload_error.error_list:
                raise
            echo.wrapped(
                "You are required to register this snap before continuing. "
                "Refer to 'snapcraft help register' for more options."
            )
            if echo.confirm(
                "Would you like to register {!r} with the Snap Store?".format(snap_name)
            ):
                self.register(snap_name=snap_name)
                return method(self, *args, snap_name=snap_name, **kwargs)
            else:
                raise

    return register_decorator


class StoreClientCLI(storeapi.StoreClient):
    """A CLI friendly StoreClient implementation."""

    # While not complete, the goal of this class is to orchestrate calls
    # to the storeapi and provide a friendly CLI interface around those
    # calls with regards to authentication and authorization, spinners or
    # progressbars and credential configuration.
    #
    # Methods defined here shall also be folded into this class as new
    # features are developed for them, but still provide a simple wrapper
    # method around those methods for backwards compatibility.
    #
    # This class can be thought of and extension to snapcraft.cli.store.
    # It just lives in snapcraft._store due to the convenience of the
    # methods it is trying to replace. Considering this is a private module
    # and this class is not exported, moving it to snapcraft.cli can take
    # place.
    #
    # This is the list of items that needs to be tackled to get to there:
    #
    # TODO create an internal copy of snapcraft.storeapi
    # TODO move configuration loading to this class and out of
    #      snapcraft.storeapi.StoreClient
    # TODO Move progressbar implementation out of snapcraft.storeapi used
    #      during upload into this class using click.
    # TODO use an instance of this class directly from snapcraft.cli.store

    @_login_wrapper
    def close_channels(
        self, *, snap_id: str, channel_names: List[str]
    ) -> Dict[str, Any]:
        return super().close_channels(snap_id=snap_id, channel_names=channel_names)

    @_login_wrapper
    def get_snap_releases(self, *, snap_name: str) -> "Releases":
        return super().get_snap_releases(snap_name=snap_name)

    @_login_wrapper
    def get_snap_channel_map(self, *, snap_name: str) -> "ChannelMap":
        return super().get_snap_channel_map(snap_name=snap_name)

    @_login_wrapper
    def get_account_information(self) -> Dict[str, Any]:
        return super().get_account_information()

    @_login_wrapper
    @_register_wrapper
    def upload_precheck(self, *, snap_name: str) -> None:
        return super().upload_precheck(snap_name=snap_name)

    @_login_wrapper
    @_register_wrapper
    def upload_metadata(
        self, *, snap_name: str, metadata: Dict[str, str], force: bool
    ) -> Dict[str, Any]:
        return super().upload_metadata(
            snap_name=snap_name, metadata=metadata, force=force
        )

    @_login_wrapper
    def register(
        self,
        *,
        snap_name: str,
        is_private: bool = False,
        store_id: Optional[str] = None,
    ) -> None:
        super().register(snap_name=snap_name, is_private=is_private, store_id=store_id)

    @_login_wrapper
    def release(
        self,
        *,
        snap_name: str,
        revision: str,
        channels: List[str],
        progressive_percentage: Optional[int] = None,
    ) -> Dict[str, Any]:
        return super().release(
            snap_name=snap_name,
            revision=revision,
            channels=channels,
            progressive_percentage=progressive_percentage,
        )

    @_login_wrapper
    @_register_wrapper
    def upload(
        self,
        *,
        snap_name: str,
        snap_filename: str,
        built_at: Optional[str] = None,
        channels: Optional[List[str]] = None,
        delta_format: Optional[str] = None,
        source_hash: Optional[str] = None,
        target_hash: Optional[str] = None,
        delta_hash: Optional[str] = None,
    ) -> "StatusTracker":
        return super().upload(
            snap_name=snap_name,
            snap_filename=snap_filename,
            built_at=built_at,
            channels=channels,
            delta_format=delta_format,
            source_hash=source_hash,
            target_hash=target_hash,
            delta_hash=delta_hash,
        )


def list_registered():
    account_info = StoreClientCLI().get_account_information()
    snaps = [
        (
            name,
            info["since"],
            "private" if info["private"] else "public",
            info["price"] or "-",
            "-",
        )
        for name, info in account_info["snaps"].get(DEFAULT_SERIES, {}).items()
        # Presenting only approved snap registrations, which means name
        # disputes will be displayed/sorted some other way.
        if info["status"] == "Approved"
    ]

    if not snaps:
        echo.warning("There are no registered snaps.")
        return

    tabulated_snaps = tabulate(
        sorted(snaps, key=operator.itemgetter(0)),
        headers=["Name", "Since", "Visibility", "Price", "Notes"],
        tablefmt="plain",
    )
    print(tabulated_snaps)


def _get_usable_keys(name=None):
    snap_path = get_host_tool_path(command_name="snap", package_name="snapd")
    keys = json.loads(
        subprocess.check_output([snap_path, "keys", "--json"], universal_newlines=True)
    )
    if keys is not None:
        for key in keys:
            if name is None or name == key["name"]:
                yield key


def _select_key(keys):
    if len(keys) > 1:
        print("Select a key:")
        print()
        tabulated_keys = tabulate(
            [(i + 1, key["name"], key["sha3-384"]) for i, key in enumerate(keys)],
            headers=["Number", "Name", "SHA3-384 fingerprint"],
            tablefmt="plain",
        )
        print(tabulated_keys)
        print()
        while True:
            try:
                keynum = int(echo.prompt("Key number: ")) - 1
            except ValueError:
                continue
            if keynum >= 0 and keynum < len(keys):
                return keys[keynum]
    else:
        return keys[0]


def _export_key(name, account_id):
    return subprocess.check_output(
        ["snap", "export-key", "--account={}".format(account_id), name],
        universal_newlines=True,
    )


def list_keys():
    keys = list(_get_usable_keys())
    account_info = StoreClientCLI().get_account_information()
    enabled_keys = {
        account_key["public-key-sha3-384"]
        for account_key in account_info["account_keys"]
    }
    if keys and enabled_keys:
        tabulated_keys = tabulate(
            [
                (
                    "*" if key["sha3-384"] in enabled_keys else "-",
                    key["name"],
                    key["sha3-384"],
                    "" if key["sha3-384"] in enabled_keys else "(not registered)",
                )
                for key in keys
            ],
            headers=["", "Name", "SHA3-384 fingerprint", ""],
            tablefmt="plain",
        )
        print(tabulated_keys)
    elif not keys and enabled_keys:
        registered_keys = "\n".join([f"- {key}" for key in enabled_keys])
        print(
            "No keys have been created on this system. "
            " See 'snapcraft create-key --help' to create a key.\n"
            "The following SHA3-384 key fingerprints have been registered "
            f"but are not available on this system:\n{registered_keys}"
        )
    else:
        print(
            "No keys have been registered."
            " See 'snapcraft register-key --help' to register a key."
        )


def create_key(name):
    if not name:
        name = "default"
    keys = list(_get_usable_keys(name=name))
    if keys:
        # `snap create-key` would eventually fail, but we can save the user
        # some time in this obvious error case by not bothering to talk to
        # the store first.
        raise storeapi.errors.KeyAlreadyRegisteredError(name)
    try:
        account_info = StoreClientCLI().get_account_information()
        enabled_names = {
            account_key["name"] for account_key in account_info["account_keys"]
        }
    except storeapi.http_clients.errors.InvalidCredentialsError:
        # Don't require a login here; if they don't have valid credentials,
        # then they probably also don't have a key registered with the store
        # yet.
        enabled_names = set()
    if name in enabled_names:
        raise storeapi.errors.KeyAlreadyRegisteredError(name)
    subprocess.check_call(["snap", "create-key", name])


def _maybe_prompt_for_key(name):
    keys = list(_get_usable_keys(name=name))
    if not keys:
        if name is not None:
            raise storeapi.errors.NoSuchKeyError(name)
        else:
            raise storeapi.errors.NoKeysError
    return _select_key(keys)


def register_key(name, use_candid: bool = False) -> None:
    key = _maybe_prompt_for_key(name)
    store_client = StoreClientCLI(use_candid=use_candid)

    # TODO: remove coupling.
    if isinstance(store_client.auth_client, storeapi.http_clients.CandidClient):
        store_client.login(acls=["modify_account_key"], save=False)
    else:
        login(store=store_client, acls=["modify_account_key"], save=False)

    logger.info("Registering key ...")
    account_info = store_client.get_account_information()
    account_key_request = _export_key(key["name"], account_info["account_id"])
    store_client.register_key(account_key_request)
    logger.info(
        'Done. The key "{}" ({}) may be used to sign your assertions.'.format(
            key["name"], key["sha3-384"]
        )
    )


def register(snap_name: str, is_private: bool = False, store_id: str = None) -> None:
    logger.info("Registering {}.".format(snap_name))
    StoreClientCLI().register(
        snap_name=snap_name, is_private=is_private, store_id=store_id
    )


def _generate_snap_build(authority_id, snap_id, grade, key_name, snap_filename):
    """Return the signed snap-build declaration for a snap on disk."""
    snap_path = get_host_tool_path(command_name="snap", package_name="snapd")
    cmd = [
        snap_path,
        "sign-build",
        "--developer-id=" + authority_id,
        "--snap-id=" + snap_id,
        "--grade=" + grade,
    ]
    if key_name:
        cmd.extend(["-k", key_name])
    cmd.append(snap_filename)
    try:
        return subprocess.check_output(cmd)
    except subprocess.CalledProcessError as e:
        raise storeapi.errors.SignBuildAssertionError(snap_filename) from e


def sign_build(snap_filename, key_name=None, local=False):
    if not os.path.exists(snap_filename):
        raise FileNotFoundError("The file {!r} does not exist.".format(snap_filename))

    snap_yaml = _get_data_from_snap_file(snap_filename)
    snap_name = snap_yaml["name"]
    grade = snap_yaml.get("grade", "stable")

    store_client = StoreClientCLI()
    account_info = store_client.get_account_information()

    try:
        authority_id = account_info["account_id"]
        snap_id = account_info["snaps"][DEFAULT_SERIES][snap_name]["snap-id"]
    except KeyError as e:
        raise storeapi.errors.StoreBuildAssertionPermissionError(
            snap_name, DEFAULT_SERIES
        ) from e

    snap_build_path = snap_filename + "-build"
    if os.path.isfile(snap_build_path):
        logger.info("A signed build assertion for this snap already exists.")
        with open(snap_build_path, "rb") as fd:
            snap_build_content = fd.read()
    else:
        key = _maybe_prompt_for_key(key_name)
        if not local:
            is_registered = [
                a
                for a in account_info["account_keys"]
                if a["public-key-sha3-384"] == key["sha3-384"]
            ]
            if not is_registered:
                raise storeapi.errors.KeyNotRegisteredError(key["name"])
        snap_build_content = _generate_snap_build(
            authority_id, snap_id, grade, key["name"], snap_filename
        )
        with open(snap_build_path, "w+") as fd:
            fd.write(snap_build_content.decode())
        logger.info("Build assertion {} saved to disk.".format(snap_build_path))

    if not local:
        store_client.push_snap_build(snap_id, snap_build_content.decode())
        logger.info("Build assertion {} uploaded to the Store.".format(snap_build_path))


def upload_metadata(snap_filename, force):
    """Upload only the metadata to the server.

    If force=True it will force the local metadata into the Store,
    ignoring any possible conflict.
    """
    logger.debug("Uploading metadata to the Store (force=%s)", force)

    # get the metadata from the snap
    snap_yaml = _get_data_from_snap_file(snap_filename)
    metadata = {
        "summary": snap_yaml["summary"],
        "description": snap_yaml["description"],
    }

    # followed by the non mandatory keys
    if "license" in snap_yaml:
        metadata["license"] = snap_yaml["license"]

    if "title" in snap_yaml:
        metadata["title"] = snap_yaml["title"]

    # other snap info
    snap_name = snap_yaml["name"]

    # hit the server
    store_client = StoreClientCLI()
    store_client.upload_precheck(snap_name=snap_name)
    store_client.upload_metadata(snap_name=snap_name, metadata=metadata, force=force)
    with _get_icon_from_snap_file(snap_filename) as icon:
        metadata = {"icon": icon}
        store_client.upload_binary_metadata(snap_name, metadata, force)

    logger.info("The metadata has been uploaded")


def upload(snap_filename, release_channels=None) -> Tuple[str, int]:
    """Upload a snap_filename to the store.

    If a cached snap is available, a delta will be generated from
    the cached snap to the new target snap and uploaded instead. In the
    case of a delta processing or upload failure, upload will fall back to
    uploading the full snap.

    If release_channels is defined it also releases it to those channels if the
    store deems the uploaded snap as ready to release.
    """
    snap_yaml = _get_data_from_snap_file(snap_filename)
    snap_name = snap_yaml["name"]
    built_at = snap_yaml.get("snapcraft-started-at")

    logger.debug(
        "Run upload precheck and verify cached data for {!r}.".format(snap_filename)
    )
    store_client = StoreClientCLI()
    store_client.upload_precheck(snap_name=snap_name)

    snap_cache = cache.SnapCache(project_name=snap_name)

    try:
        deb_arch = snap_yaml["architectures"][0]
    except KeyError:
        deb_arch = "all"

    source_snap = snap_cache.get(deb_arch=deb_arch)
    sha3_384_available = hasattr(hashlib, "sha3_384")

    result: Optional[Dict[str, Any]] = None
    if sha3_384_available and source_snap:
        try:
            result = _upload_delta(
                store_client,
                snap_name=snap_name,
                snap_filename=snap_filename,
                source_snap=source_snap,
                built_at=built_at,
                channels=release_channels,
            )
        except storeapi.errors.StoreDeltaApplicationError as e:
            logger.warning(
                "Error generating delta: {}\n"
                "Falling back to uploading full snap...".format(str(e))
            )
        except storeapi.errors.StoreUploadError as upload_error:
            logger.warning(
                "Unable to upload delta to store: {}\n"
                "Falling back to uploading full snap...".format(upload_error.error_list)
            )

    if result is None:
        result = _upload_snap(
            store_client,
            snap_name=snap_name,
            snap_filename=snap_filename,
            built_at=built_at,
            channels=release_channels,
        )

    snap_cache.cache(snap_filename=snap_filename)
    snap_cache.prune(deb_arch=deb_arch, keep_hash=calculate_sha3_384(snap_filename))

    return snap_name, result["revision"]


def _upload_snap(
    store_client,
    *,
    snap_name: str,
    snap_filename: str,
    built_at: str,
    channels: Optional[List[str]],
) -> Dict[str, Any]:
    tracker = store_client.upload(
        snap_name=snap_name,
        snap_filename=snap_filename,
        built_at=built_at,
        channels=channels,
    )
    result = tracker.track()
    tracker.raise_for_code()
    return result


def _upload_delta(
    store_client,
    *,
    snap_name: str,
    snap_filename: str,
    source_snap: str,
    built_at: str,
    channels: Optional[List[str]] = None,
) -> Dict[str, Any]:
    delta_format = "xdelta3"
    logger.debug("Found cached source snap {}.".format(source_snap))
    target_snap = os.path.join(os.getcwd(), snap_filename)

    try:
        xdelta_generator = deltas.XDelta3Generator(
            source_path=source_snap, target_path=target_snap
        )
        delta_filename = xdelta_generator.make_delta()
    except (DeltaGenerationError, DeltaGenerationTooBigError, ToolMissingError) as e:
        raise storeapi.errors.StoreDeltaApplicationError(str(e))

    snap_hashes = {
        "source_hash": calculate_sha3_384(source_snap),
        "target_hash": calculate_sha3_384(target_snap),
        "delta_hash": calculate_sha3_384(delta_filename),
    }

    try:
        logger.debug("Uploading delta {!r}.".format(delta_filename))
        delta_tracker = store_client.upload(
            snap_name=snap_name,
            snap_filename=delta_filename,
            built_at=built_at,
            channels=channels,
            delta_format=delta_format,
            source_hash=snap_hashes["source_hash"],
            target_hash=snap_hashes["target_hash"],
            delta_hash=snap_hashes["delta_hash"],
        )
        result = delta_tracker.track()
        delta_tracker.raise_for_code()
    except storeapi.errors.StoreReviewError as e:
        if e.code == "processing_upload_delta_error":
            raise storeapi.errors.StoreDeltaApplicationError(str(e))
        else:
            raise
    except storeapi.http_clients.errors.StoreServerError as e:
        raise storeapi.errors.StoreUploadError(snap_name, e.response)
    finally:
        if os.path.isfile(delta_filename):
            try:
                os.remove(delta_filename)
            except OSError:
                logger.warning("Unable to remove delta {}.".format(delta_filename))
    return result


def _get_text_for_opened_channels(opened_channels):
    if len(opened_channels) == 1:
        return "The {!r} channel is now open.".format(opened_channels[0])
    else:
        channels = ("{!r}".format(channel) for channel in opened_channels[:-1])
        return "The {} and {!r} channels are now open.".format(
            ", ".join(channels), opened_channels[-1]
        )


def _get_text_for_channel(channel):
    if "progressive" in channel:
        notes = "progressive ({}%)".format(channel["progressive"]["percentage"])
    else:
        notes = "-"

    if channel["info"] == "none":
        channel_text = (channel["channel"], "-", "-", notes, "")
    elif channel["info"] == "tracking":
        channel_text = (channel["channel"], "^", "^", notes, "")
    elif channel["info"] == "specific":
        channel_text = (
            channel["channel"],
            channel["version"],
            channel["revision"],
            notes,
            "",
        )
    elif channel["info"] == "branch":
        channel_text = (
            channel["channel"],
            channel["version"],
            channel["revision"],
            notes,
            channel["expires_at"],
        )
    else:
        logger.error(
            "Unexpected channel info: %r in channel %s",
            channel["info"],
            channel["channel"],
        )
        channel_text = (channel["channel"], "", "", "", "")

    return channel_text


def _tabulated_channel_map_tree(channel_map_tree):
    """Tabulate channel map (LTS Channel channel-maps)"""

    def _format_tree(channel_maps, track):
        arches = []

        for arch, channel_map in sorted(channel_maps.items()):
            arches += [
                (printable_arch,) + _get_text_for_channel(channel)
                for (printable_arch, channel) in zip(
                    [arch] + [""] * len(channel_map), channel_map
                )
            ]

        return [
            (printable_arch,) + printable_track
            for (printable_arch, printable_track) in zip(
                [track] + [""] * len(arches), arches
            )
        ]

    data = []
    for track, track_data in sorted(channel_map_tree.items()):
        channel_maps = {}
        for series, series_data in track_data.items():
            for arch, channel_map in series_data.items():
                channel_maps[arch] = channel_map
        parsed_channels = [channel for channel in _format_tree(channel_maps, track)]
        data += parsed_channels

    have_expiration = any(x[6] for x in data)
    expires_at_header = "Expires at" if have_expiration else ""
    headers = [
        "Track",
        "Arch",
        "Channel",
        "Version",
        "Revision",
        "Notes",
        expires_at_header,
    ]
    return tabulate(data, numalign="left", headers=headers, tablefmt="plain")


def download(
    snap_name,
    *,
    arch: str,
    download_path: str,
    risk: str,
    track: Optional[str] = None,
    except_hash="",
) -> str:
    """Download snap from the store to download_path.
    :param str snap_name: The snap name to download.
    :param str risk: the channel risk get the snap from.
    :param str track: the specific channel track get the snap from.
    :param str download_path: the path to write the downloaded snap to.
    :param str arch: the architecture of the download as a deb arch.
    :param str except_hash: do not download if set to a sha3_384 hash that
                            matches the snap_name to be downloaded.
    :raises storeapi.errors.SHAMismatchErrorRuntimeError:
         If the checksum for the downloaded file does not match the expected
         hash.
    :returns: A sha3_384 of the file that was or would have been downloaded.
    """
    return StoreClientCLI().download(
        snap_name,
        risk=risk,
        track=track,
        download_path=download_path,
        arch=arch,
        except_hash=except_hash,
    )


def status(snap_name, arch):
    status = StoreClientCLI().get_snap_status(snap_name, arch)

    channel_map_tree = status.get("channel_map_tree", {})
    # This does not look good in green so we print instead
    tabulated_status = _tabulated_channel_map_tree(channel_map_tree)
    print(tabulated_status)


def gated(snap_name):
    """Print list of snaps gated by snap_name."""
    store_client = StoreClientCLI()
    account_info = store_client.get_account_information()
    # Get data for the gating snap
    snaps = account_info.get("snaps", {})

    # Resolve name to snap-id
    try:
        snap_id = snaps[DEFAULT_SERIES][snap_name]["snap-id"]
    except KeyError:
        raise storeapi.errors.SnapNotFoundError(snap_name=snap_name)

    validations = store_client.get_assertion(snap_id, endpoint="validations")

    if validations:
        table_data = []
        for v in validations:
            name = v["approved-snap-name"]
            revision = v["approved-snap-revision"]
            if revision == "-":
                revision = None
            required = str(v.get("required", True))
            # Currently timestamps have microseconds, which look bad
            timestamp = v["timestamp"]
            if "." in timestamp:
                timestamp = timestamp.split(".")[0] + "Z"
            table_data.append([name, revision, required, timestamp])
        tabulated = tabulate(
            table_data,
            headers=["Name", "Revision", "Required", "Approved"],
            tablefmt="plain",
            missingval="-",
        )
        print(tabulated)
    else:
        print("There are no validations for snap {!r}".format(snap_name))


def validate(
    snap_name: str,
    validations: List[str],
    revoke: bool = False,
    key: Optional[str] = None,
):
    """Generate, sign and upload validation assertions."""
    # Check validations format
    _check_validations(validations)

    # Need the ID of the logged in user.
    store_client = StoreClientCLI()
    account_info = store_client.get_account_information()
    authority_id = account_info["account_id"]

    # Get data for the gating snap
    try:
        snap_id = account_info["snaps"][DEFAULT_SERIES][snap_name]["snap-id"]
    except KeyError:
        raise storeapi.errors.SnapNotFoundError(snap_name=snap_name)

    existing_validations = {
        (i["approved-snap-id"], i["approved-snap-revision"]): i
        for i in store_client.get_assertion(
            snap_id, endpoint="validations", params={"include_revoked": "1"}
        )
    }

    # Then, for each requested validation, generate assertion
    for validation in validations:
        gated_snap, rev = validation.split("=", 1)
        echo.info(f"Getting details for {gated_snap}")
        # The Info API is not authed, so it cannot see private snaps.
        try:
            approved_data = store_client.snap.get_info(gated_snap)
            approved_snap_id = approved_data.snap_id
        except storeapi.errors.SnapNotFoundError:
            approved_snap_id = gated_snap

        assertion_payload = {
            "type": "validation",
            "authority-id": authority_id,
            "series": DEFAULT_SERIES,
            "snap-id": snap_id,
            "approved-snap-id": approved_snap_id,
            "approved-snap-revision": rev,
            "timestamp": datetime.utcnow().isoformat() + "Z",
            "revoked": "true" if revoke else "false",
        }

        # check for existing validation assertions
        existing = existing_validations.get((approved_snap_id, rev))
        if existing:
            previous_revision = int(existing.get("revision", "0"))
            assertion_payload["revision"] = str(previous_revision + 1)

        assertion = _sign_assertion(validation, assertion_payload, key, "validations")

        # Save assertion to a properly named file
        fname = f"{snap_name}-{gated_snap}-r{rev}.assertion"
        with open(fname, "wb") as f:
            f.write(assertion)

        store_client.push_assertion(snap_id, assertion, endpoint="validations")


validation_re = re.compile("^[^=]+=[0-9]+$")


def _check_validations(validations):
    invalids = [v for v in validations if not validation_re.match(v)]
    if invalids:
        raise storeapi.errors.InvalidValidationRequestsError(invalids)


def _sign_assertion(snap_name, assertion, key, endpoint):
    cmdline = ["snap", "sign"]
    if key:
        cmdline += ["-k", key]
    snap_sign = Popen(
        cmdline, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    data = json.dumps(assertion).encode("utf8")
    echo.info("Signing {} assertion for {}".format(endpoint, snap_name))
    assertion, err = snap_sign.communicate(input=data)
    if snap_sign.returncode != 0:
        err = err.decode()
        raise storeapi.errors.StoreAssertionError(
            endpoint=endpoint, snap_name=snap_name, error=err
        )

    return assertion



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

from snapcraft.internal import project_loader
from snapcraft.project import Project as _Project
from tests import unit


class ProjectLoaderBaseTest(unit.TestCase):
    def make_snapcraft_project(self, snapcraft_yaml, project_kwargs=None):
        snapcraft_yaml_file_path = self.make_snapcraft_yaml(snapcraft_yaml)
        if project_kwargs is None:
            project_kwargs = dict()
        project = _Project(
            snapcraft_yaml_file_path=snapcraft_yaml_file_path, **project_kwargs
        )
        return project_loader.load_config(project)


class LoadPartBaseTest(ProjectLoaderBaseTest):
    def setUp(self):
        super().setUp()

        patcher = mock.patch(
            "snapcraft.internal.project_loader._parts_config.PartsConfig.load_part"
        )
        self.mock_load_part = patcher.start()
        self.addCleanup(patcher.stop)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from unittest import mock

import fixtures
from testtools.matchers import Contains, Equals

from snapcraft import storeapi
from snapcraft.cli import assertions
from tests import unit

from . import StoreCommandsBaseTestCase


class CollaborateBaseTestCase(StoreCommandsBaseTestCase):
    def setUp(self):
        super().setUp()
        patcher = mock.patch("subprocess.Popen")
        self.popen_mock = patcher.start()
        process_mock = mock.Mock()
        process_mock.returncode = 0
        process_mock.communicate.return_value = [b"foo", b""]
        self.popen_mock.return_value = process_mock
        self.addCleanup(patcher.stop)
        patcher = mock.patch("snapcraft.cli.assertions._edit_developers")
        self.edit_developers_mock = patcher.start()
        self.edit_developers_mock.return_value = [
            {"developer-id": "dummy-id", "since": "2015-07-19 19:30:00"}
        ]
        self.addCleanup(patcher.stop)
        self.client.login(email="dummy", password="test correct password")


class CollaboratorsCommandTestCase(CollaborateBaseTestCase):

    command_name = "edit-collaborators"

    def test_collaborators_without_snap_name_must_error(self):
        result = self.run_command([self.command_name])

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(result.output, Contains("Usage:"))

    def test_collaborators_update_developers(self):
        result = self.run_command(
            [self.command_name, "test-snap-with-dev", "--key-name", "key-name"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.popen_mock.assert_called_with(
            ["snap", "sign", "-k", "key-name"], stderr=-1, stdin=-1, stdout=-1
        )

    def test_collaborators_add_developers(self):
        result = self.run_command([self.command_name, "core", "--key-name", "key-name"])

        self.assertThat(result.exit_code, Equals(0))
        self.popen_mock.assert_called_with(
            ["snap", "sign", "-k", "key-name"], stderr=-1, stdin=-1, stdout=-1
        )

    def test_no_prior_developer_assertion(self):
        result = self.run_command(
            [self.command_name, "core-no-dev", "--key-name", "key-name"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.popen_mock.assert_called_with(
            ["snap", "sign", "-k", "key-name"], stderr=-1, stdin=-1, stdout=-1
        )

    def test_collaborate_no_revoke_uploads_request(self):
        result = self.run_command(
            [self.command_name] + ["no-revoked", "--key-name", "key-name"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains("This will revoke the following collaborators: 'this'"),
        )
        self.assertThat(
            result.output, Contains("Are you sure you want to continue? [y/N]:")
        )
        self.assertThat(
            result.output,
            Contains("The collaborators for this snap have not been altered."),
        )

    def test_collaborate_unchanged_collaborators(self):
        self.edit_developers_mock.return_value = [
            {
                "developer-id": "test-dev-id",
                "since": "2017-02-10 08:35:00",
                "until": "2018-02-10 08:35:00",
            }
        ]

        result = self.run_command(
            [self.command_name] + ["test-snap-with-dev", "--key-name", "key-name"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output, Contains("Aborting due to unchanged collaborators list.")
        )


class CollaborateErrorsTestCase(CollaborateBaseTestCase):

    command_name = "edit-collaborators"

    def test_collaborate_snap_not_found(self):
        raised = self.assertRaises(
            storeapi.errors.SnapNotFoundError,
            self.run_command,
            [self.command_name, "notfound", "--key-name", "key-name"],
        )

        self.assertThat(str(raised), Contains("Snap 'notfound' was not found"))

    def test_collaborate_bad_request(self):
        raised = self.assertRaises(
            storeapi.errors.StoreValidationError,
            self.run_command,
            [self.command_name, "badrequest", "--key-name", "key-name"],
        )

        self.assertThat(
            str(raised),
            Contains(
                "Received error 400: 'The given `snap-id` does not match "
                "the assertion.'"
            ),
        )

    def test_collaborate_yes_revoke_uploads_request(self):
        raised = self.assertRaises(
            storeapi.errors.StoreValidationError,
            self.run_command,
            [self.command_name, "revoked", "--key-name", "key-name"],
            input="y\n",
        )

        self.assertThat(
            str(raised),
            Contains(
                'Received error 409: "{}"'.format(
                    "The assertion's `developers` would revoke existing uploads."
                )
            ),
        )


class EditDevelopersTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        patcher = mock.patch("subprocess.check_call")
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_edit_developers_must_write_header_and_developers(self):
        developers_from_assertion = [
            {
                "developer-id": "test-dev-id1",
                "since": "2017-02-10T08:35:00.000000Z",
                "until": "2018-02-10T08:35:00.000000Z",
            },
            {
                "developer-id": "test-dev-id2",
                "since": "2016-02-10T08:35:00.000000Z",
                "until": "2019-02-10T08:35:00.000000Z",
            },
        ]

        existing_developers = (
            "developers:\n"
            "- developer-id: test-dev-id1\n"
            "  since: '2017-02-10 08:35:00'\n"
            "  until: '2018-02-10 08:35:00'\n"
            "- developer-id: test-dev-id2\n"
            "  since: '2016-02-10 08:35:00'\n"
            "  until: '2019-02-10 08:35:00'\n"
        )
        expected_written = assertions._COLLABORATION_HEADER + "\n" + existing_developers

        with mock.patch(
            "builtins.open", new_callable=mock.mock_open, read_data=expected_written
        ) as mock_open:
            developers_for_assertion = assertions._update_developers(
                developers_from_assertion
            )

        written = ""
        for call in mock_open().write.call_args_list:
            written += str(call.call_list()[0][0][0])
        self.assertThat(written, Equals(expected_written))
        self.assertThat(developers_for_assertion, Equals(developers_from_assertion))

    def test_edit_developers_must_return_new_developers(self):
        developers_from_assertion = [
            {
                "developer-id": "test-dev-id1",
                "since": "2017-02-10T08:35:00.000000Z",
                "until": "2018-02-10T08:35:00.000000Z",
            }
        ]

        new_developers = (
            "developers:\n"
            "- developer-id: test-dev-id1\n"
            "  since: '2017-02-10 08:35:00'\n"
            "  until: '2018-02-10 08:35:00'\n"
            "- developer-id: test-dev-id2\n"
            "  since: '2016-02-10 08:35:00'\n"
            "  until: '2019-02-10 08:35:00'\n"
        )

        with mock.patch(
            "builtins.open", new_callable=mock.mock_open, read_data=new_developers
        ):
            developers_for_assertion = assertions._update_developers(
                developers_from_assertion
            )

        expected_developers = developers_from_assertion + [
            {
                "developer-id": "test-dev-id2",
                "since": "2016-02-10T08:35:00.000000Z",
                "until": "2019-02-10T08:35:00.000000Z",
            }
        ]
        self.assertThat(developers_for_assertion, Equals(expected_developers))


class EditDevelopersOpenEditorTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        patcher = mock.patch("subprocess.check_call")
        self.check_call_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def test_edit_collaborators_must_open_editor_default(self):
        self.useFixture(fixtures.EnvironmentVariable("EDITOR", None))
        assertions._edit_developers({})
        self.check_call_mock.assert_called_with(["vi", mock.ANY])

    def test_edit_collaborators_must_open_editor_environment(self):
        self.useFixture(fixtures.EnvironmentVariable("EDITOR", "test-editor"))
        assertions._edit_developers({})
        self.check_call_mock.assert_called_with(["test-editor", mock.ANY])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict, List, Optional, Set

import jsonschema

from ._api_schema import CHANNEL_MAP_JSONSCHEMA

"""
This module holds representations for results for the v2 channel-map
API endpoint provided by the Snap Store.

The full API is documented on
https://dashboard.snapcraft.io/docs/v2/en/snaps.html#snap-channel-map
"""


class Progressive:
    """Represent Progressive information for a MappedChannel."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "Progressive":
        jsonschema.validate(
            payload,
            CHANNEL_MAP_JSONSCHEMA["properties"]["channel-map"]["items"]["properties"][
                "progressive"
            ],
        )
        return cls(
            paused=payload["paused"],
            percentage=payload["percentage"],
            current_percentage=payload["current-percentage"],
        )

    def marshal(self) -> Dict[str, Any]:
        return {
            "paused": self.paused,
            "percentage": self.percentage,
            "current-percentage": self.current_percentage,
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.current_percentage!r}=>{self.percentage!r}>"

    def __init__(
        self,
        *,
        paused: Optional[bool],
        percentage: Optional[float],
        current_percentage: Optional[float],
    ) -> None:
        self.paused = paused
        self.percentage = percentage
        self.current_percentage = current_percentage


class MappedChannel:
    """Represent a mapped channel item for "channel-map" in ChannelMap."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "MappedChannel":
        jsonschema.validate(
            payload, CHANNEL_MAP_JSONSCHEMA["properties"]["channel-map"]["items"]
        )
        return cls(
            channel=payload["channel"],
            revision=payload["revision"],
            architecture=payload["architecture"],
            expiration_date=payload["expiration-date"],
            progressive=Progressive.unmarshal(payload["progressive"]),
        )

    def marshal(self) -> Dict[str, Any]:
        return {
            "channel": self.channel,
            "revision": self.revision,
            "architecture": self.architecture,
            "expiration-date": self.expiration_date,
            "progressive": self.progressive.marshal(),
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.channel!r} for revision {self.revision!r} and architecture {self.architecture!r}>"

    def __init__(
        self,
        *,
        channel: str,
        revision: int,
        architecture: str,
        expiration_date: Optional[str],
        progressive: Progressive,
    ) -> None:
        self.channel = channel
        self.revision = revision
        self.architecture = architecture
        self.expiration_date = expiration_date
        self.progressive = progressive


class Revision:
    """Represent a revision item for "revisions" in ChannelMap."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "Revision":
        jsonschema.validate(
            payload, CHANNEL_MAP_JSONSCHEMA["properties"]["revisions"]["items"]
        )
        return cls(
            revision=payload["revision"],
            version=payload["version"],
            architectures=payload["architectures"],
        )

    def marshal(self) -> Dict[str, Any]:
        return {
            "revision": self.revision,
            "version": self.version,
            "architectures": self.architectures,
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.revision!r} for version {self.version!r} and architectures {self.architectures!r}>"

    def __init__(
        self, *, revision: int, version: str, architectures: List[str]
    ) -> None:
        self.revision = revision
        self.version = version
        self.architectures = architectures


class SnapChannel:
    """Represent a channel item in "channels" in Snap."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "SnapChannel":
        jsonschema.validate(
            payload,
            CHANNEL_MAP_JSONSCHEMA["properties"]["snap"]["properties"]["channels"][
                "items"
            ],
        )
        return cls(
            name=payload["name"],
            track=payload["track"],
            risk=payload["risk"],
            branch=payload["branch"],
            fallback=payload["fallback"],
        )

    def marshal(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "track": self.track,
            "risk": self.risk,
            "branch": self.branch,
            "fallback": self.fallback,
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.name!r}>"

    def __init__(
        self,
        *,
        name: str,
        track: str,
        risk: str,
        branch: Optional[str],
        fallback: Optional[str],
    ) -> None:
        self.name = name
        self.track = track
        self.risk = risk
        self.branch = branch
        self.fallback = fallback


class SnapTrack:
    """Represent a track item in "tracks" in Snap."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "SnapTrack":
        jsonschema.validate(
            payload,
            CHANNEL_MAP_JSONSCHEMA["properties"]["snap"]["properties"]["tracks"][
                "items"
            ],
        )
        return cls(
            name=payload["name"],
            status=payload["status"],
            creation_date=payload["creation-date"],
            version_pattern=payload["version-pattern"],
        )

    def marshal(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "status": self.status,
            "creation-date": self.creation_date,
            "version-pattern": self.version_pattern,
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.name!r}>"

    def __init__(
        self,
        *,
        name: str,
        status: str,
        creation_date: Optional[str],
        version_pattern: Optional[str],
    ) -> None:
        self.name = name
        self.status = status
        self.creation_date = creation_date
        self.version_pattern = version_pattern


class Snap:
    """Represent "snap" in ChannelMap."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "Snap":
        jsonschema.validate(payload, CHANNEL_MAP_JSONSCHEMA["properties"]["snap"])
        return cls(
            name=payload["name"],
            channels=[SnapChannel.unmarshal(sc) for sc in payload["channels"]],
            tracks=[SnapTrack.unmarshal(st) for st in payload["tracks"]],
        )

    def marshal(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "channels": [sc.marshal() for sc in self.channels],
            "tracks": [st.marshal() for st in self.tracks],
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.name!r}>"

    def __init__(
        self, *, name: str, channels: List[SnapChannel], tracks: List[SnapTrack]
    ) -> None:
        self.name = name
        self.channels = channels
        self.tracks = tracks


class ChannelMap:
    """Represent the data returned from the channel-map Snap Store endpoint."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "ChannelMap":
        jsonschema.validate(payload, CHANNEL_MAP_JSONSCHEMA)
        return cls(
            channel_map=[MappedChannel.unmarshal(c) for c in payload["channel-map"]],
            revisions=[Revision.unmarshal(r) for r in payload["revisions"]],
            snap=Snap.unmarshal(payload["snap"]),
        )

    def marshal(self) -> Dict[str, Any]:
        return {
            "channel-map": [c.marshal() for c in self.channel_map],
            "revisions": [r.marshal() for r in self.revisions],
            "snap": self.snap.marshal(),
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.snap.name!r}>"

    def __init__(
        self, *, channel_map: List[MappedChannel], revisions: List[Revision], snap: Snap
    ) -> None:
        self.channel_map = channel_map
        self.revisions = revisions
        self.snap = snap

    def get_mapped_channel(
        self, *, channel_name: str, architecture: str, progressive: bool
    ) -> MappedChannel:
        channels_with_name = (
            cm for cm in self.channel_map if cm.channel == channel_name
        )
        channels_with_arch = (
            cm for cm in channels_with_name if cm.architecture == architecture
        )

        if progressive:
            channels = [
                cm for cm in channels_with_arch if cm.progressive.percentage is not None
            ]
        else:
            channels = [
                cm for cm in channels_with_arch if cm.progressive.percentage is None
            ]

        try:
            return channels[0]
        except IndexError:
            raise ValueError(
                f"No channel mapped to {channel_name!r} for architecture {architecture!r} when progressive is {progressive!r}"
            )

    def get_channel_info(self, channel_name: str) -> SnapChannel:
        for snap_channel in self.snap.channels:
            if snap_channel.name == channel_name:
                return snap_channel
        raise ValueError(f"No channel information for {channel_name!r}")

    def get_revision(self, revision_number: int) -> Revision:
        for revision_item in self.revisions:
            if revision_item.revision == revision_number:
                return revision_item
        raise ValueError(f"No revision information for {revision_number!r}")

    def get_existing_architectures(self) -> Set[str]:
        architectures: List[str] = list()
        for revision_item in self.revisions:
            architectures.extend(revision_item.architectures)

        return set(architectures)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

import snapcraft.plugins.v2._ros as _ros
import snapcraft.plugins.v2.catkin_tools as catkin_tools


def test_schema():
    assert catkin_tools.CatkinToolsPlugin.get_schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "additionalProperties": False,
        "properties": {
            "catkin-tools-cmake-args": {
                "default": [],
                "items": {"type": "string"},
                "minItems": 0,
                "type": "array",
            },
            "catkin-tools-packages": {
                "items": {"type": "string"},
                "minItems": 0,
                "type": "array",
                "uniqueItems": True,
            },
        },
        "type": "object",
    }


def test_get_build_packages():
    plugin = catkin_tools.CatkinToolsPlugin(part_name="my-part", options=lambda: None)

    assert plugin.get_build_packages() == {
        "python3-rosdep",
        "python3-catkin-tools",
        "python3-osrf-pycommon",
    }


def test_get_build_environment():
    plugin = catkin_tools.CatkinToolsPlugin(part_name="my-part", options=lambda: None)

    assert plugin.get_build_environment() == {
        "ROS_PYTHON_VERSION": "3",
    }


def test_out_of_source_build_property():
    plugin = catkin_tools.CatkinToolsPlugin(part_name="my-part", options=lambda: None)

    assert plugin.out_of_source_build


def test_get_build_commands(monkeypatch):
    class Options:
        catkin_tools_cmake_args = list()
        catkin_tools_packages = list()

    plugin = catkin_tools.CatkinToolsPlugin(part_name="my-part", options=Options())

    monkeypatch.setattr(sys, "path", ["", "/test"])
    monkeypatch.setattr(sys, "executable", "/test/python3")
    monkeypatch.setattr(_ros, "__file__", "/test/_ros.py")
    monkeypatch.setattr(os, "environ", dict())

    assert plugin.get_build_commands() == [
        'state="$(set +o)"',
        "set +u",
        "_CATKIN_SETUP_DIR=/opt/ros/$ROS_DISTRO . /opt/ros/$ROS_DISTRO/setup.sh",
        'eval "$(state)"',
        "if [ ! -f /etc/ros/rosdep/sources.list.d/20-default.list ]; then sudo rosdep "
        "init; fi",
        "rosdep update --include-eol-distros --rosdistro $ROS_DISTRO",
        "rosdep install --default-yes --ignore-packages-from-source --from-paths $SNAPCRAFT_PART_SRC",
        "catkin init",
        "catkin profile add -f default",
        "catkin config --profile default --install "
        "--source-space $SNAPCRAFT_PART_SRC --build-space $SNAPCRAFT_PART_BUILD "
        "--install-space $SNAPCRAFT_PART_INSTALL/opt/ros/$ROS_DISTRO",
        "catkin build --no-notify --profile default -j $SNAPCRAFT_PARALLEL_BUILD_COUNT",
        "env -i LANG=C.UTF-8 LC_ALL=C.UTF-8 /test/python3 -I "
        "/test/_ros.py "
        "stage-runtime-dependencies --part-src $SNAPCRAFT_PART_SRC --part-install $SNAPCRAFT_PART_INSTALL "
        "--ros-distro $ROS_DISTRO --target-arch $SNAPCRAFT_TARGET_ARCH",
    ]


def test_get_build_commands_with_all_properties(monkeypatch):
    class Options:
        catkin_tools_cmake_args = ["cmake", "args..."]
        catkin_tools_packages = ["package1", "package2..."]

    plugin = catkin_tools.CatkinToolsPlugin(part_name="my-part", options=Options())

    monkeypatch.setattr(sys, "path", ["", "/test"])
    monkeypatch.setattr(sys, "executable", "/test/python3")
    monkeypatch.setattr(_ros, "__file__", "/test/_ros.py")
    monkeypatch.setattr(
        os,
        "environ",
        dict(
            FOO="baR",
            PATH="/bin:/test",
            SNAP="TESTSNAP",
            SNAP_ARCH="TESTARCH",
            SNAP_NAME="TESTSNAPNAME",
            SNAP_VERSION="TESTV1",
            http_proxy="http://foo",
            https_proxy="https://bar",
        ),
    )

    assert plugin.get_build_commands() == [
        'state="$(set +o)"',
        "set +u",
        "_CATKIN_SETUP_DIR=/opt/ros/$ROS_DISTRO . /opt/ros/$ROS_DISTRO/setup.sh",
        'eval "$(state)"',
        "if [ ! -f /etc/ros/rosdep/sources.list.d/20-default.list ]; then sudo rosdep "
        "init; fi",
        "rosdep update --include-eol-distros --rosdistro $ROS_DISTRO",
        "rosdep install --default-yes --ignore-packages-from-source --from-paths $SNAPCRAFT_PART_SRC",
        "catkin init",
        "catkin profile add -f default",
        "catkin config --profile default --install "
        "--source-space $SNAPCRAFT_PART_SRC --build-space $SNAPCRAFT_PART_BUILD "
        "--install-space $SNAPCRAFT_PART_INSTALL/opt/ros/$ROS_DISTRO --cmake-args cmake args...",
        "catkin build --no-notify --profile default -j $SNAPCRAFT_PARALLEL_BUILD_COUNT package1 package2...",
        "env -i LANG=C.UTF-8 LC_ALL=C.UTF-8 PATH=/bin:/test SNAP=TESTSNAP "
        "SNAP_ARCH=TESTARCH SNAP_NAME=TESTSNAPNAME SNAP_VERSION=TESTV1 "
        "http_proxy=http://foo https_proxy=https://bar "
        "/test/python3 -I "
        "/test/_ros.py "
        "stage-runtime-dependencies --part-src $SNAPCRAFT_PART_SRC --part-install $SNAPCRAFT_PART_INSTALL "
        "--ros-distro $ROS_DISTRO --target-arch $SNAPCRAFT_TARGET_ARCH",
    ]



from setuptools import find_packages, setup

setup(
    name="test",
    version="0.1",
    packages=find_packages(),
    entry_points={
        "console_scripts": ["python3_test = python3_test_package.__main__:main"]
    },
)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import subprocess

import testtools


class HasArchitecture:
    """Match if the file was built for the expected architecture"""

    def __init__(self, expected_arch):
        self._expected_arch = expected_arch

    def __str__(self):
        return "HasArchitecture()"

    def match(self, file_path):
        cmd = ["file", file_path]
        try:
            output = subprocess.check_output(cmd).decode()
        except subprocess.CalledProcessError as cmd_err:
            raise ValueError(
                "Failed to get architecture info: {}", cmd_err
            ) from cmd_err

        try:
            # output is of the form
            # ELF 64-bit LSB executable, ARM aarch64, version 1 (SYSV), \
            # dynamically linked, interpreter /lib/ld-linux-aarch64.so.1, \
            # for GNU/Linux 3.7.0, \
            # BuildID[sha1]=58b8495bef7a6c9e5f403879e29dc63ad4ed2d11, \
            # not stripped
            arch = output.split(",")[1]
        except IndexError as e:
            raise ValueError("Failed to parse magic {!r}".format(output)) from e
        if self._expected_arch not in arch:
            return testtools.matchers.Mismatch(
                "Expected {!r} to be in {!r}".format(self._expected_arch, arch)
            )



#!/usr/bin/python3

# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import logging
import os
import subprocess
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import cpu_count

import yaml

# path for shellcheck binary
SHELLCHECK = os.getenv("SHELLCHECK", "shellcheck")
# set to non-empty to ignore all errors
NO_FAIL = os.getenv("NO_FAIL")
# set to non empty to enable 'set -x'
D = os.getenv("D")
# set to non-empty to enable verbose logging
V = os.getenv("V")
# set to a number to use these many threads
N = int(os.getenv("N") or cpu_count())
# file with list of files that can fail validation
CAN_FAIL = os.getenv("CAN_FAIL")

# names of sections
SECTIONS = frozenset(
    [
        "prepare",
        "prepare-each",
        "restore",
        "restore-each",
        "debug",
        "debug-each",
        "execute",
        "repack",
    ]
)


def parse_arguments():
    parser = argparse.ArgumentParser(description="spread shellcheck helper")
    parser.add_argument("-s", "--shell", default="bash", help="shell")
    parser.add_argument(
        "-n",
        "--no-errors",
        action="store_true",
        default=False,
        help="ignore all errors ",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", default=False, help="verbose logging"
    )
    parser.add_argument(
        "--can-fail",
        default=None,
        help=("file with list of files that are can fail " "validation"),
    )
    parser.add_argument(
        "-P",
        "--max-procs",
        default=N,
        type=int,
        metavar="N",
        help="run these many shellchecks in parallel (default: %(default)s)",
    )
    parser.add_argument("paths", nargs="+", help="paths to check")
    return parser.parse_args()


class ShellcheckRunError(Exception):
    def __init__(self, stderr):
        super().__init__()
        self.stderr = stderr


class ShellcheckError(Exception):
    def __init__(self, path):
        super().__init__()
        self.sectionerrors = {}
        self.path = path

    def addfailure(self, section, error):
        self.sectionerrors[section] = error

    def __len__(self):
        return len(self.sectionerrors)


class ShellcheckFailures(Exception):
    def __init__(self, failures=None):
        super().__init__()
        self.failures = set()
        if failures:
            self.failures = set(failures)

    def merge(self, otherfailures):
        self.failures = self.failures.union(otherfailures.failures)

    def __len__(self):
        return len(self.failures)

    def intersection(self, other):
        return self.failures.intersection(other)

    def difference(self, other):
        return self.failures.difference(other)

    def __iter__(self):
        return iter(self.failures)


def checksection(data):
    # spread shell snippets are executed under 'set -e' shell, make sure
    # shellcheck knows about that
    data = "set -eu\n" + data
    proc = subprocess.Popen(
        [SHELLCHECK, "-s", "bash", "-x", "-"],
        stdout=subprocess.PIPE,
        stdin=subprocess.PIPE,
    )
    stdout, _ = proc.communicate(input=data.encode("utf-8"), timeout=10)
    if proc.returncode != 0:
        raise ShellcheckRunError(stdout)


def checkfile(path):
    logging.debug("checking file %s", path)
    with open(path) as inf:
        data = yaml.load(inf)

    errors = ShellcheckError(path)

    for section in SECTIONS & data.keys():
        try:
            logging.debug("%s: checking section %s", path, section)
            checksection(data[section])
        except ShellcheckRunError as serr:
            errors.addfailure(section, serr.stderr.decode("utf-8"))

    if path.endswith("spread.yaml") and "suites" in data:
        # check suites
        for suite_name, suite in data["suites"].items():
            for section in SECTIONS & suite.keys():
                try:
                    logging.debug(
                        "%s (suite %s): checking section %s", path, suite_name, section
                    )
                    checksection(suite[section])
                except ShellcheckRunError as serr:
                    errors.addfailure(
                        "suites/" + suite_name + "/" + section,
                        serr.stderr.decode("utf-8"),
                    )

    if errors:
        raise errors


def findfiles(indir):
    for root, _, files in os.walk(indir, topdown=True):
        for name in files:
            if name in ["spread.yaml", "task.yaml"]:
                yield os.path.join(root, name)


def checkpath(loc, max_workers):
    if os.path.isdir(loc):
        # setup iterator
        locations = findfiles(loc)
    else:
        locations = [loc]

    failed = []

    def check1path(path):
        try:
            checkfile(path)
        except ShellcheckError as err:
            return err
        return None

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        for serr in executor.map(check1path, locations):
            if serr is None:
                continue
            logging.error(
                ("shellcheck failed for file %s in sections: " "%s; error log follows"),
                serr.path,
                ", ".join(serr.sectionerrors.keys()),
            )
            for section, error in serr.sectionerrors.items():
                logging.error("%s: section '%s':\n%s", serr.path, section, error)
            failed.append(serr.path)

    if failed:
        raise ShellcheckFailures(failures=failed)


def loadfilelist(flistpath):
    flist = set()
    with open(flistpath) as inf:
        for line in inf:
            if not line.startswith("#"):
                flist.add(line.strip())
    return flist


def main(opts):
    paths = opts.paths or ["."]
    failures = ShellcheckFailures()
    for pth in paths:
        try:
            checkpath(pth, opts.max_procs)
        except ShellcheckFailures as sf:
            failures.merge(sf)

    if failures:
        if opts.can_fail:
            can_fail = loadfilelist(opts.can_fail)

            unexpected = failures.difference(can_fail)
            if unexpected:
                logging.error(
                    (
                        "validation failed for the following "
                        "non-whitelisted files:\n%s"
                    ),
                    "\n".join([" - " + f for f in sorted(unexpected)]),
                )
                raise SystemExit(1)

            did_not_fail = can_fail - failures.intersection(can_fail)
            if did_not_fail:
                logging.error(
                    (
                        "the following files are whitelisted "
                        "but validated successfully:\n%s"
                    ),
                    "\n".join([" - " + f for f in sorted(did_not_fail)]),
                )
                raise SystemExit(1)

            # no unexpected failures
            return

        logging.error(
            "validation failed for the following files:\n%s",
            "\n".join([" - " + f for f in sorted(failures)]),
        )

        if NO_FAIL or opts.no_errors:
            logging.warning("ignoring errors")
        else:
            raise SystemExit(1)


if __name__ == "__main__":
    opts = parse_arguments()
    if opts.verbose or D or V:
        subprocess.call([SHELLCHECK, "--version"])
        lvl = logging.DEBUG
    else:
        lvl = logging.INFO
    logging.basicConfig(level=lvl)

    if CAN_FAIL:
        opts.can_fail = CAN_FAIL

    if NO_FAIL:
        opts.no_errors = True

    main(opts)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015, 2017-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from testtools.matchers import Equals

from snapcraft.plugins.v1.nil import NilPlugin
from tests import unit


class TestNilPlugin(unit.TestCase):
    def test_schema(self):
        schema = NilPlugin.schema()
        self.assertThat(
            schema["$schema"], Equals("http://json-schema.org/draft-04/schema#")
        )
        self.assertThat(schema["type"], Equals("object"))
        self.assertFalse(schema["additionalProperties"])
        self.assertFalse(schema["properties"])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import tarfile
import threading

import fixtures
from pyftpdlib import authorizers, handlers, servers
from testtools.matchers import DirExists, FileContains, FileExists

from tests import integration


class FtpServerRunning(fixtures.Fixture):
    def __init__(self, directory):
        super().__init__()
        self.directory = directory

    def setUp(self):
        super().setUp()
        self._start_ftp_server()

    def _start_ftp_server(self):
        authorizer = authorizers.DummyAuthorizer()
        authorizer.add_anonymous(self.directory)
        handler = handlers.FTPHandler
        handler.authorizer = authorizer
        self.server = servers.FTPServer(("", 2121), handler)

        server_thread = threading.Thread(target=self.server.serve_forever)
        server_thread.start()
        self.addCleanup(self._stop_ftp_server, server_thread)

    def _stop_ftp_server(self, thread):
        self.server.close_all()
        thread.join()


class DumpPluginTestCase(integration.TestCase):
    def test_stage_dump_plugin(self):
        self.run_snapcraft("stage", "dump")

        expected_files = [
            "flat",
            os.path.join("flatdir", "flat2"),
            "onedeep",
            os.path.join("onedeepdir", "onedeep2"),
            "oneflat",
            "top-simple",
            "notop",
            "parent",
            "slash",
            "readonly_file",
        ]
        for expected_file in expected_files:
            self.assertThat(os.path.join(self.stage_dir, expected_file), FileExists())
        expected_dirs = ["dir-simple", "notopdir"]
        for expected_dir in expected_dirs:
            self.assertThat(os.path.join(self.stage_dir, expected_dir), DirExists())

        # Regression test for
        # https://bugs.launchpad.net/snapcraft/+bug/1500728
        self.run_snapcraft("pull")

    def test_download_file_with_content_encoding_set(self):
        """Download a file with Content-Encoding: gzip LP: #1611776"""
        self.run_snapcraft("pull", "compressed-content-encoding")

    def test_download_file_from_ftp_source(self):
        """Download a file from a FTP source, LP: #1602323"""
        ftp_dir = os.path.join(self.path, "ftp")
        os.mkdir(ftp_dir)
        ftp_server = FtpServerRunning(ftp_dir)
        self.useFixture(ftp_server)

        test_file_path = os.path.join(self.path, "test")
        with open(test_file_path, "w") as test_file:
            test_file.write("Hello ftp")
        with tarfile.open(os.path.join(ftp_dir, "test.tar.gz"), "w:gz") as tar:
            tar.add(test_file_path)

        self.run_snapcraft("pull", "ftp-source")
        self.assertThat(
            os.path.join(self.parts_dir, "ftp-part", "src", "test"),
            FileContains("Hello ftp"),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The cmake plugin is useful for building cmake based parts.

These are projects that have a CMakeLists.txt that drives the build.
The plugin requires a CMakeLists.txt in the root of the source tree.

If the part has a list of build-snaps listed, the part will be set up in
such a way that the paths to those snaps are used as paths for find_package
and find_library by use of `CMAKE_FIND_ROOT_PATH``.
This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - configflags:
      (list of strings)
      configure flags to pass to the build using the common cmake semantics.
"""

import logging
import os
from typing import List, Optional

from snapcraft.plugins.v1 import PluginV1

logger = logging.getLogger(name=__name__)


class _Flag:
    def __str__(self) -> str:
        if self.value is None:
            flag = self.name
        else:
            flag = "{}={}".format(self.name, self.value)
        return flag

    def __init__(self, flag: str) -> None:
        parts = flag.split("=")
        self.name = parts[0]
        try:
            self.value: Optional[str] = parts[1]
        except IndexError:
            self.value = None


class CMakePlugin(PluginV1):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["configflags"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        # For backwards compatibility
        schema["properties"]["make-parameters"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return super().get_build_properties() + ["configflags"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)
        self.build_packages.append("cmake")
        self.out_of_source_build = True

        if options.make_parameters:
            logger.warning("make-paramaters is deprecated, ignoring.")

    def build(self):
        source_subdir = getattr(self.options, "source_subdir", None)
        if source_subdir:
            sourcedir = os.path.join(self.sourcedir, source_subdir)
        else:
            sourcedir = self.sourcedir

        env = self._build_environment()
        configflags = self._get_processed_flags()

        self.run(["cmake", sourcedir, "-DCMAKE_INSTALL_PREFIX="] + configflags, env=env)

        # TODO: there is a better way to specify the job count on newer versions of cmake
        # https://github.com/Kitware/CMake/commit/1ab3881ec9e809ac5f6cad5cd84048310b8683e2
        self.run(
            ["cmake", "--build", ".", "--", "-j{}".format(self.parallel_build_count)],
            env=env,
        )

        self.run(["cmake", "--build", ".", "--target", "install"], env=env)

    def _get_processed_flags(self) -> List[str]:
        # Return the original if no build_snaps are in options.
        if not self.options.build_snaps:
            return self.options.configflags

        build_snap_paths = [
            os.path.join(os.path.sep, "snap", snap_name.split("/")[0], "current")
            for snap_name in self.options.build_snaps
        ]

        flags = [_Flag(f) for f in self.options.configflags]
        for flag in flags:
            if flag.name == "-DCMAKE_FIND_ROOT_PATH":
                flag.value = "{};{}".format(flag.value, ";".join(build_snap_paths))
                break
        else:
            flags.append(
                _Flag("-DCMAKE_FIND_ROOT_PATH={}".format(";".join(build_snap_paths)))
            )

        return [str(f) for f in flags]

    def _build_environment(self):
        env = os.environ.copy()
        env["DESTDIR"] = self.installdir
        env["CMAKE_PREFIX_PATH"] = "$CMAKE_PREFIX_PATH:{}".format(
            self.project.stage_dir
        )
        env["CMAKE_INCLUDE_PATH"] = "$CMAKE_INCLUDE_PATH:" + ":".join(
            ["{0}/include", "{0}/usr/include", "{0}/include/{1}", "{0}/usr/include/{1}"]
        ).format(self.project.stage_dir, self.project.arch_triplet)
        env["CMAKE_LIBRARY_PATH"] = "$CMAKE_LIBRARY_PATH:" + ":".join(
            ["{0}/lib", "{0}/usr/lib", "{0}/lib/{1}", "{0}/usr/lib/{1}"]
        ).format(self.project.stage_dir, self.project.arch_triplet)

        return env



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import os
from unittest import mock

import jsonschema
from testtools.matchers import Equals, HasLength

from snapcraft.internal import errors
from snapcraft.plugins.v1 import python
from tests import fixture_setup, unit

from . import PluginsV1BaseTestCase


def setup_directories(plugin, python_version, create_setup_py=True):
    version = "2.7" if python_version == "python2" else "3.5"
    os.makedirs(plugin.sourcedir)
    os.makedirs(plugin.builddir)
    python_home = os.path.join(plugin.installdir, "usr")
    python_lib_path = os.path.join(python_home, "lib", "python" + version)
    python_include_path = os.path.join(python_home, "include", "python" + version)

    os.makedirs(os.path.join(python_lib_path, "dist-packages"))
    os.makedirs(python_include_path)
    if create_setup_py:
        open(os.path.join(plugin.sourcedir, "setup.py"), "w").close()

    site_path = os.path.join(
        plugin.installdir, "lib", "python" + version, "site-packages"
    )
    os.makedirs(site_path)
    with open(os.path.join(python_lib_path, "site.py"), "w") as f:
        f.write(
            "#!/usr/bin/python3\n"
            "# comment\n"
            "ENABLE_USER_SITE = None\n"
            "USER_SITE = None\n"
            "USER_BASE = None\n"
            "# comment\n"
        )


class PythonPluginBaseTest(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class Options:
            source = "."
            source_subdir = ""
            requirements = []
            constraints = ""
            python_version = "python3"
            python_packages = []
            process_dependency_links = False

        self.options = Options()

        patcher = mock.patch("snapcraft.plugins.v1._python.Pip")
        self.mock_pip = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch.object(python.PythonPlugin, "_setup_tools_install")
        self.mock_setup_tools = patcher.start()
        self.addCleanup(patcher.stop)


class PythonPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        schema = python.PythonPlugin.schema()
        expected_requirements = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        expected_constraints = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        expected_python_packages = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        expected_process_dependency_links = {"type": "boolean", "default": False}
        expected_python_version = {
            "type": "string",
            "default": "python3",
            "enum": ["python2", "python3"],
        }

        self.assertDictEqual(
            expected_requirements, schema["properties"]["requirements"]
        )
        self.assertDictEqual(expected_constraints, schema["properties"]["constraints"])
        self.assertDictEqual(
            expected_python_packages, schema["properties"]["python-packages"]
        )
        self.assertDictEqual(
            expected_process_dependency_links,
            schema["properties"]["process-dependency-links"],
        )
        self.assertDictEqual(
            expected_python_version, schema["properties"]["python-version"]
        )

    def test_get_pull_properties(self):
        expected_pull_properties = [
            "requirements",
            "constraints",
            "python-packages",
            "process-dependency-links",
            "python-version",
        ]
        resulting_pull_properties = python.PythonPlugin.get_pull_properties()

        self.assertThat(
            resulting_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, resulting_pull_properties)


class PythonPluginSchemaValidationTest(unit.TestCase):
    def test_sources_validation_neither(self):
        schema = self._get_schema()
        properties = {}
        self.assertRaises(
            jsonschema.ValidationError, jsonschema.validate, properties, schema
        )

    def test_sources_validation_source(self):
        schema = self._get_schema()
        properties = {"source": ""}
        jsonschema.validate(properties, schema)

    def test_sources_validation_packages(self):
        schema = self._get_schema()
        properties = {"python-packages": []}
        jsonschema.validate(properties, schema)

    def test_sources_validation_both(self):
        schema = self._get_schema()
        properties = {"source": "", "python-packages": []}
        jsonschema.validate(properties, schema)

    def _get_schema(self):
        schema = python.PythonPlugin.schema()
        # source definition comes from the main schema
        schema["properties"]["source"] = {"type": "string"}
        return schema


class PythonPluginTest(PythonPluginBaseTest):
    def test_env(self):
        plugin = python.PythonPlugin("test-part", self.options, self.project)
        expected_env = []
        env = plugin.env("/testpath")
        self.assertListEqual(expected_env, env)

        env_missing_path = plugin.env("/testpath")
        self.assertTrue("PYTHONPATH=/testpath" not in env_missing_path)

    def test_pull_with_setup_py(self):
        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)

        plugin.pull()

        pip_download = self.mock_pip.return_value.download
        pip_download.assert_called_once_with(
            [],
            constraints=set(),
            process_dependency_links=False,
            requirements=set(),
            setup_py_dir=None,
        )

        self.mock_pip.return_value.wheel.assert_not_called()
        self.mock_pip.return_value.install.assert_not_called()

    def test_pull_with_requirements(self):
        self.options.requirements = ["requirements.txt"]
        self.options.python_packages = ["test", "packages"]

        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)

        requirements_path = os.path.join(plugin.sourcedir, "requirements.txt")
        open(requirements_path, "w").close()

        plugin.pull()

        pip_download = self.mock_pip.return_value.download
        pip_download.assert_called_once_with(
            ["test", "packages"],
            constraints=set(),
            process_dependency_links=False,
            requirements={requirements_path},
            setup_py_dir=None,
        )

        self.mock_pip.return_value.wheel.assert_not_called()
        self.mock_pip.return_value.install.assert_not_called()

    def test_pull_with_constraints(self):
        self.options.constraints = ["constraints.txt"]
        self.options.python_packages = ["test", "packages"]

        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)

        constraints_path = os.path.join(plugin.sourcedir, "constraints.txt")
        open(constraints_path, "w").close()

        plugin.pull()

        pip_download = self.mock_pip.return_value.download
        pip_download.assert_called_once_with(
            ["test", "packages"],
            constraints={constraints_path},
            process_dependency_links=False,
            requirements=set(),
            setup_py_dir=None,
        )

        self.mock_pip.return_value.wheel.assert_not_called()
        self.mock_pip.return_value.install.assert_not_called()

    @mock.patch.object(python.PluginV1, "build")
    def test_build(self, mock_base_build):
        self.options.requirements = ["requirements.txt"]
        self.options.constraints = ["constraints.txt"]
        self.options.python_packages = ["test", "packages"]

        packages = collections.OrderedDict()
        packages["yaml"] = "1.2"
        packages["extras"] = "1.0"
        self.mock_pip.return_value.list.return_value = packages

        self.useFixture(fixture_setup.CleanEnvironment())
        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)

        for file_name in self.options.requirements + self.options.constraints:
            path = os.path.join(plugin.sourcedir, file_name)
            open(path, "w").close()

        requirements_path = os.path.join(plugin.builddir, "requirements.txt")
        constraints_path = os.path.join(plugin.builddir, "constraints.txt")

        def build_side_effect():
            open(os.path.join(plugin.builddir, "setup.py"), "w").close()
            os.mkdir(os.path.join(plugin.builddir, "dist"))
            open(os.path.join(plugin.builddir, "dist", "package.tar"), "w").close()
            open(requirements_path, "w").close()
            open(constraints_path, "w").close()

        mock_base_build.side_effect = build_side_effect

        pip_wheel = self.mock_pip.return_value.wheel
        pip_wheel.return_value = ["foo", "bar"]

        plugin.build()

        # Pip should not attempt to download again in build (only pull)
        pip_download = self.mock_pip.return_value.download
        pip_download.assert_not_called()

        pip_wheel.assert_called_once_with(
            ["test", "packages"],
            constraints={constraints_path},
            process_dependency_links=False,
            requirements={requirements_path},
            setup_py_dir=None,
        )

        pip_install = self.mock_pip.return_value.install
        self.assertThat(pip_install.call_count, Equals(2))
        pip_install.assert_has_calls(
            [
                mock.call(
                    ["foo", "bar"],
                    process_dependency_links=False,
                    upgrade=True,
                    install_deps=False,
                ),
                mock.call(
                    [],
                    setup_py_dir=plugin.builddir,
                    constraints={constraints_path},
                    process_dependency_links=False,
                    upgrade=True,
                ),
            ]
        )

    def test_pip_with_url(self):
        self.options.requirements = ["https://test.com/requirements.txt"]
        self.options.constraints = ["http://test.com/constraints.txt"]

        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)

        # Patch requests so we don't hit the network when the requirements
        # and constraints files are downloaded to save their contents in the
        # manifest.
        with mock.patch("requests.get"):
            plugin.pull()
            plugin.build()

        pip_download = self.mock_pip.return_value.download
        pip_download.assert_called_once_with(
            [],
            constraints=set(self.options.constraints),
            process_dependency_links=False,
            requirements=set(self.options.requirements),
            setup_py_dir=None,
        )

        pip_install = self.mock_pip.return_value.install
        self.assertThat(pip_install.call_count, Equals(2))
        pip_install.assert_has_calls(
            [
                mock.call(
                    [], upgrade=True, process_dependency_links=False, install_deps=False
                ),
                mock.call(
                    [],
                    setup_py_dir=plugin.sourcedir,
                    constraints=set(self.options.constraints),
                    process_dependency_links=False,
                    upgrade=True,
                ),
            ]
        )

        pip_wheel = self.mock_pip.return_value.wheel
        pip_wheel.assert_called_once_with(
            [],
            constraints=set(self.options.constraints),
            process_dependency_links=False,
            requirements=set(self.options.requirements),
            setup_py_dir=None,
        )

    def test_fileset_ignores(self):
        plugin = python.PythonPlugin("test-part", self.options, self.project)
        expected_fileset = [
            "-bin/pip",
            "-bin/pip2",
            "-bin/pip3",
            "-bin/pip2.7",
            "-bin/pip3.*",
            "-bin/easy_install*",
            "-bin/wheel",
            "-**/__pycache__",
            "-**/*.pyc",
            "-lib/python*/site-packages/*/RECORD",
        ]
        fileset = plugin.snap_fileset()
        self.assertListEqual(expected_fileset, fileset)

    def test_process_dependency_links(self):
        self.options.process_dependency_links = True
        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)
        plugin.pull()
        plugin.build()

        pip_download = self.mock_pip.return_value.download
        pip_download.assert_called_once_with(
            [],
            constraints=set(),
            process_dependency_links=True,
            requirements=set(),
            setup_py_dir=None,
        )

        pip_install = self.mock_pip.return_value.install
        self.assertThat(pip_install.call_count, Equals(2))
        pip_install.assert_has_calls(
            [
                mock.call(
                    [], upgrade=True, process_dependency_links=True, install_deps=False
                ),
                mock.call(
                    [],
                    setup_py_dir=plugin.sourcedir,
                    constraints=set(),
                    process_dependency_links=True,
                    upgrade=True,
                ),
            ]
        )

        pip_wheel = self.mock_pip.return_value.wheel
        pip_wheel.assert_called_once_with(
            [],
            constraints=set(),
            process_dependency_links=True,
            requirements=set(),
            setup_py_dir=None,
        )

    def test_get_manifest_with_python_packages(self):
        packages = collections.OrderedDict()
        packages["testpackage1"] = "1.0"
        packages["testpackage2"] = "1.2"
        self.mock_pip.return_value.list.return_value = packages

        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)
        plugin.build()
        self.assertThat(
            plugin.get_manifest(),
            Equals(
                collections.OrderedDict(
                    {"python-packages": ["testpackage1=1.0", "testpackage2=1.2"]}
                )
            ),
        )

    def test_get_manifest_with_local_requirements(self):
        self.options.requirements = ["requirements.txt"]
        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)
        requirements_path = os.path.join(plugin.sourcedir, "requirements.txt")
        with open(requirements_path, "w") as requirements_file:
            requirements_file.write("testpackage1==1.0\n")
            requirements_file.write("testpackage2==1.2")

        plugin.build()

        self.assertThat(
            plugin.get_manifest()["requirements-contents"],
            Equals(["testpackage1==1.0", "testpackage2==1.2"]),
        )

    def test_get_manifest_with_multiple_local_requirements(self):
        self.options.requirements = ["requirements1.txt", "requirements2.txt"]
        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)
        requirements1_path = os.path.join(plugin.sourcedir, "requirements1.txt")
        requirements2_path = os.path.join(plugin.sourcedir, "requirements2.txt")
        with open(requirements1_path, "w") as requirements_file:
            requirements_file.write("testpackage1==1.0\n")
        with open(requirements2_path, "w") as requirements_file:
            requirements_file.write("testpackage2==1.2")

        plugin.build()

        self.assertThat(
            plugin.get_manifest()["requirements-contents"],
            Equals(["testpackage1==1.0", "testpackage2==1.2"]),
        )

    def test_get_manifest_with_local_constraints(self):
        self.options.constraints = ["constraints.txt"]

        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)
        constraints_path = os.path.join(plugin.sourcedir, "constraints.txt")
        with open(constraints_path, "w") as constraints_file:
            constraints_file.write("testpackage1==1.0\n")
            constraints_file.write("testpackage2==1.2")

        plugin.build()

        self.assertThat(
            plugin.get_manifest()["constraints-contents"],
            Equals(["testpackage1==1.0", "testpackage2==1.2"]),
        )

    def test_get_manifest_with_multiple_local_constraints(self):
        self.options.constraints = ["constraints1.txt", "constraints2.txt"]

        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)
        constraints1_path = os.path.join(plugin.sourcedir, "constraints1.txt")
        with open(constraints1_path, "w") as constraints_file:
            constraints_file.write("testpackage1==1.0\n")
        constraints2_path = os.path.join(plugin.sourcedir, "constraints2.txt")
        with open(constraints2_path, "w") as constraints_file:
            constraints_file.write("testpackage2==1.2")

        plugin.build()

        self.assertThat(
            plugin.get_manifest()["constraints-contents"],
            Equals(["testpackage1==1.0", "testpackage2==1.2"]),
        )

    def test_no_python_packages_does_nothing(self):
        # This should be an error but given that we default to
        # 'source: .' and now that pip 10 has been released
        # we run into the need of fixing this situation.
        self.mock_pip.return_value.list.return_value = dict()

        self.useFixture(fixture_setup.CleanEnvironment())
        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version, create_setup_py=False)

        pip_wheel = self.mock_pip.return_value.wheel
        pip_wheel.return_value = []

        plugin.build()

        # Pip should not attempt to download again in build (only pull)
        pip_download = self.mock_pip.return_value.download
        pip_download.assert_not_called()

        pip_wheel.assert_called_once_with(
            [],
            constraints=set(),
            process_dependency_links=False,
            requirements=set(),
            setup_py_dir=None,
        )

        pip_install = self.mock_pip.return_value.install
        pip_install.assert_not_called()


class PythonCore18Test(PythonPluginBaseTest):
    def test_plugin_stage_packages_python2(self):
        self.options.python_version = "python2"

        plugin = python.PythonPlugin("test-part", self.options, self.project)
        self.assertThat(plugin.plugin_stage_packages, Equals(["python"]))

    def test_plugin_stage_packages_python3(self):
        self.options.python_version = "python3"

        plugin = python.PythonPlugin("test-part", self.options, self.project)
        self.assertThat(
            plugin.plugin_stage_packages, Equals(["python3", "python3-distutils"])
        )


class PythonCoreTest(PythonPluginBaseTest):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "core"

    def test_plugin_stage_packages_python2(self):
        self.options.python_version = "python2"

        plugin = python.PythonPlugin("test-part", self.options, self.project)
        self.assertThat(plugin.plugin_stage_packages, Equals(["python"]))

    def test_plugin_stage_packages_python3(self):
        self.options.python_version = "python3"

        plugin = python.PythonPlugin("test-part", self.options, self.project)
        self.assertThat(plugin.plugin_stage_packages, Equals(["python3"]))


class FileMissingPythonPluginTest(PythonPluginBaseTest):
    def test_constraints_file_missing(self):
        self.options.constraints = "constraints.txt"

        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)

        self.assertRaises(python.SnapcraftPluginPythonFileMissing, plugin.pull)

    def test_requirements_file_missing(self):
        self.options.requirements = "requirements.txt"

        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)

        self.assertRaises(python.SnapcraftPluginPythonFileMissing, plugin.pull)


class PythonPluginWithURLTestCase(
    PythonPluginBaseTest, unit.FakeFileHTTPServerBasedTestCase
):
    def setUp(self):
        super().setUp()
        self.source = "http://{}:{}/{}".format(
            *self.server.server_address, "testfile.txt"
        )

    def test_get_manifest_with_requirements_url(self):
        self.options.requirements = [self.source]
        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)

        plugin.build()

        self.assertThat(
            plugin.get_manifest()["requirements-contents"], Equals(["Test fake file"])
        )

    def test_get_manifest_with_constraints_url(self):
        self.options.constraints = [self.source]
        plugin = python.PythonPlugin("test-part", self.options, self.project)
        setup_directories(plugin, self.options.python_version)

        plugin.build()

        self.assertThat(
            plugin.get_manifest()["constraints-contents"], Equals(["Test fake file"])
        )


class PythonPluginUnsupportedBase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "unsupported-base"

        class Options:
            source = "dir"

        self.options = Options()

    def test_unsupported_base_raises(self):
        self.assertRaises(
            errors.PluginBaseError,
            python.PythonPlugin,
            "test-part",
            self.options,
            self.project,
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys

import snapcraft.plugins.v2._ros as _ros
import snapcraft.plugins.v2.colcon as colcon


def test_schema():
    assert colcon.ColconPlugin.get_schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "additionalProperties": False,
        "properties": {
            "colcon-ament-cmake-args": {
                "default": [],
                "items": {"type": "string"},
                "minItems": 0,
                "type": "array",
            },
            "colcon-catkin-cmake-args": {
                "default": [],
                "items": {"type": "string"},
                "minItems": 0,
                "type": "array",
            },
            "colcon-cmake-args": {
                "default": [],
                "items": {"type": "string"},
                "minItems": 0,
                "type": "array",
            },
            "colcon-packages": {
                "items": {"type": "string"},
                "minItems": 0,
                "type": "array",
                "uniqueItems": True,
            },
            "colcon-packages-ignore": {
                "default": [],
                "items": {"type": "string"},
                "minItems": 0,
                "type": "array",
                "uniqueItems": True,
            },
        },
        "type": "object",
    }


def test_get_build_packages():
    plugin = colcon.ColconPlugin(part_name="my-part", options=lambda: None)

    assert plugin.get_build_packages() == {
        "python3-colcon-common-extensions",
        "python3-rosdep",
        "python3-rosinstall",
        "python3-wstool",
    }


def test_get_build_environment():
    plugin = colcon.ColconPlugin(part_name="my-part", options=lambda: None)

    assert plugin.get_build_environment() == {
        "AMENT_PYTHON_EXECUTABLE": "/usr/bin/python3",
        "COLCON_PYTHON_EXECUTABLE": "/usr/bin/python3",
        "ROS_PYTHON_VERSION": "3",
    }


def test_out_of_source_build_property():
    plugin = colcon.ColconPlugin(part_name="my-part", options=lambda: None)

    assert plugin.out_of_source_build


def test_get_build_commands(monkeypatch):
    class Options:
        colcon_ament_cmake_args = list()
        colcon_catkin_cmake_args = list()
        colcon_cmake_args = list()
        colcon_packages = list()
        colcon_packages_ignore = list()

    plugin = colcon.ColconPlugin(part_name="my-part", options=Options())

    monkeypatch.setattr(sys, "path", ["", "/test"])
    monkeypatch.setattr(sys, "executable", "/test/python3")
    monkeypatch.setattr(_ros, "__file__", "/test/_ros.py")
    monkeypatch.setattr(os, "environ", dict())

    assert plugin.get_build_commands() == [
        'state="$(set +o)"',
        "set +u",
        ". /opt/ros/$ROS_DISTRO/setup.sh",
        'eval "$(state)"',
        "if [ ! -f /etc/ros/rosdep/sources.list.d/20-default.list ]; then sudo rosdep "
        "init; fi",
        "rosdep update --include-eol-distros --rosdistro $ROS_DISTRO",
        "rosdep install --default-yes --ignore-packages-from-source --from-paths $SNAPCRAFT_PART_SRC",
        "colcon build "
        "--base-paths $SNAPCRAFT_PART_SRC --build-base $SNAPCRAFT_PART_BUILD "
        "--merge-install --install-base $SNAPCRAFT_PART_INSTALL "
        "--parallel-workers ${SNAPCRAFT_PARALLEL_BUILD_COUNT}",
        "env -i LANG=C.UTF-8 LC_ALL=C.UTF-8 /test/python3 -I "
        "/test/_ros.py "
        "stage-runtime-dependencies --part-src $SNAPCRAFT_PART_SRC --part-install $SNAPCRAFT_PART_INSTALL "
        "--ros-distro $ROS_DISTRO --target-arch $SNAPCRAFT_TARGET_ARCH",
    ]


def test_get_build_commands_with_all_properties(monkeypatch):
    class Options:
        colcon_ament_cmake_args = ["ament", "args..."]
        colcon_catkin_cmake_args = ["catkin", "args..."]
        colcon_cmake_args = ["cmake", "args..."]
        colcon_packages = ["package1", "package2..."]
        colcon_packages_ignore = ["ipackage1", "ipackage2..."]

    plugin = colcon.ColconPlugin(part_name="my-part", options=Options())

    monkeypatch.setattr(sys, "path", ["", "/test"])
    monkeypatch.setattr(sys, "executable", "/test/python3")
    monkeypatch.setattr(_ros, "__file__", "/test/_ros.py")
    monkeypatch.setattr(
        os,
        "environ",
        dict(
            FOO="baR",
            PATH="/bin:/test",
            SNAP="TESTSNAP",
            SNAP_ARCH="TESTARCH",
            SNAP_NAME="TESTSNAPNAME",
            SNAP_VERSION="TESTV1",
            http_proxy="http://foo",
            https_proxy="https://bar",
        ),
    )

    assert plugin.get_build_commands() == [
        'state="$(set +o)"',
        "set +u",
        ". /opt/ros/$ROS_DISTRO/setup.sh",
        'eval "$(state)"',
        "if [ ! -f /etc/ros/rosdep/sources.list.d/20-default.list ]; then sudo rosdep "
        "init; fi",
        "rosdep update --include-eol-distros --rosdistro $ROS_DISTRO",
        "rosdep install --default-yes --ignore-packages-from-source --from-paths $SNAPCRAFT_PART_SRC",
        "colcon build "
        "--base-paths $SNAPCRAFT_PART_SRC --build-base $SNAPCRAFT_PART_BUILD "
        "--merge-install --install-base $SNAPCRAFT_PART_INSTALL "
        "--packages-ignore ipackage1 ipackage2... --packages-select package1 "
        "package2... --ament-cmake-args ament args... --catkin-cmake-args catkin "
        "args... --parallel-workers ${SNAPCRAFT_PARALLEL_BUILD_COUNT}",
        "env -i LANG=C.UTF-8 LC_ALL=C.UTF-8 PATH=/bin:/test SNAP=TESTSNAP "
        "SNAP_ARCH=TESTARCH SNAP_NAME=TESTSNAPNAME SNAP_VERSION=TESTV1 "
        "http_proxy=http://foo https_proxy=https://bar "
        "/test/python3 -I /test/_ros.py "
        "stage-runtime-dependencies --part-src $SNAPCRAFT_PART_SRC --part-install $SNAPCRAFT_PART_INSTALL "
        "--ros-distro $ROS_DISTRO --target-arch $SNAPCRAFT_TARGET_ARCH",
    ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

import apt
import testscenarios
from testtools.matchers import Equals

from tests import integration


class BuildPackageVersionTestCase(testscenarios.WithScenarios, integration.TestCase):

    scenarios = (
        ("global", dict(project="build-package", package="hello", part="hello")),
        (
            "local",
            dict(project="build-package-global", package="haskell-doc", part=None),
        ),
    )

    def test_build_package_gets_version(self):
        self.copy_project_to_cwd(self.project)
        version = self.set_build_package_version(
            os.path.join("snap", "snapcraft.yaml"), self.part, package=self.package
        )
        self.run_snapcraft("pull")

        with apt.Cache() as apt_cache:
            installed_version = apt_cache[self.package].candidate.version
        self.assertThat(installed_version, Equals(version))


class BuildPackageVersionErrorsTestCase(integration.TestCase):
    def test_build_package_with_invalid_version_must_fail(self):
        self.copy_project_to_cwd("build-package-global")
        self.set_build_package_version(
            os.path.join("snap", "snapcraft.yaml"),
            part=None,
            package="haskell-doc",
            version="invalid",
        )
        error = self.assertRaises(
            subprocess.CalledProcessError, self.run_snapcraft, "pull"
        )
        self.assertIn(
            "Could not find a required package in 'build-packages': "
            "haskell-doc=invalid",
            str(error.output),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from testtools.matchers import DirExists, Equals, FileExists

from tests import integration


class TarPluginTestCase(integration.TestCase):
    def test_stage_zip_source(self):
        self.copy_project_to_cwd("zip")
        self.run_snapcraft("stage")

        expected_files = ["exec", "top-simple", os.path.join("dir-simple", "sub")]
        for expected_file in expected_files:
            self.assertThat(os.path.join(self.stage_dir, expected_file), FileExists())
        self.assertThat(
            os.access(os.path.join(self.stage_dir, "exec"), os.X_OK), Equals(True)
        )
        self.assertThat(
            os.access(os.path.join(self.stage_dir, "non-unix"), os.X_OK), Equals(True)
        )
        expected_dirs = ["dir-simple", "non-unix"]
        for expected_dir in expected_dirs:
            self.assertThat(os.path.join(self.stage_dir, expected_dir), DirExists())

        # Regression test for
        # https://bugs.launchpad.net/snapcraft/+bug/1500728
        self.run_snapcraft("pull")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from testtools.matchers import Contains, Equals

import snapcraft.internal.project_loader._config as _config
from snapcraft.internal.project_loader import errors
from tests import unit

from . import LoadPartBaseTest, ProjectLoaderBaseTest


class VariableExpansionTest(LoadPartBaseTest):
    def test_version_script(self):
        project_config = self.make_snapcraft_project(
            dedent(
                """\
            name: test
            base: core18
            version: "1.0"
            version-script: echo $SNAPCRAFT_PROJECT_VERSION-$SNAPCRAFT_PROJECT_GRADE
            summary: test
            description: nothing
            architectures: ['amd64']
            confinement: strict
            grade: devel

            parts:
              part1:
                plugin: go
                stage-packages: [fswebcam]
        """
            )
        )

        self.assertThat(project_config.data["version-script"], Equals("echo 1.0-devel"))

    def test_config_expands_filesets(self):
        self.make_snapcraft_project(
            dedent(
                """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            grade: stable

            parts:
              part1:
                plugin: go
                stage-packages: [fswebcam]
                filesets:
                  wget:
                    - /usr/lib/wget.so
                    - /usr/bin/wget
                  build-wget:
                    - /usr/lib/wget.a
                stage:
                  - $wget
                  - $build-wget
                prime:
                  - $wget
                  - /usr/share/my-icon.png
        """
            )
        )

        self.mock_load_part.assert_called_with(
            "part1",
            "go",
            {
                "prime": [
                    "/usr/lib/wget.so",
                    "/usr/bin/wget",
                    "/usr/share/my-icon.png",
                ],
                "plugin": "go",
                "stage-packages": ["fswebcam"],
                "stage": ["/usr/lib/wget.so", "/usr/bin/wget", "/usr/lib/wget.a"],
            },
        )

    def test_replace_snapcraft_variables(self):
        self.make_snapcraft_project(
            dedent(
                """\
            name: project-name
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict

            parts:
              main:
                plugin: make
                source: $SNAPCRAFT_PROJECT_NAME-$SNAPCRAFT_PROJECT_VERSION
                make-options: [DEP=$SNAPCRAFT_STAGE]
            """
            )
        )

        self.mock_load_part.assert_called_with(
            "main",
            "make",
            {
                "source": "project-name-1",
                "plugin": "make",
                "stage": [],
                "prime": [],
                "make-options": ["DEP={}".format(self.stage_dir)],
            },
        )


class DependenciesTest(ProjectLoaderBaseTest):
    def setUp(self):
        super().setUp()

        self.config = self.make_snapcraft_project(
            dedent(
                dedent(
                    """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            grade: stable

            parts:
                main:
                    plugin: nil

                dependent:
                    plugin: nil
                    after: [main]

                nested-dependent:
                    plugin: nil
                    after: [dependent]"""
                )
            )
        )

    def assert_part_names(self, part_names, parts):
        self.assertThat({p.name for p in parts}, Equals(part_names))

    def test_get_dependencies(self):
        self.assertFalse(self.config.parts.get_dependencies("main"))
        self.assert_part_names(
            {"main"}, self.config.parts.get_dependencies("dependent")
        )
        self.assert_part_names(
            {"dependent"}, self.config.parts.get_dependencies("nested-dependent")
        )

    def test_get_dependencies_recursive(self):
        self.assertFalse(self.config.parts.get_dependencies("main", recursive=True))
        self.assert_part_names(
            {"main"}, self.config.parts.get_dependencies("dependent", recursive=True)
        )
        self.assert_part_names(
            {"main", "dependent"},
            self.config.parts.get_dependencies("nested-dependent", recursive=True),
        )

    def test_get_reverse_dependencies(self):
        self.assertFalse(self.config.parts.get_reverse_dependencies("nested-dependent"))
        self.assert_part_names(
            {"nested-dependent"},
            self.config.parts.get_reverse_dependencies("dependent"),
        )
        self.assert_part_names(
            {"dependent"}, self.config.parts.get_reverse_dependencies("main")
        )

    def test_get_reverse_dependencies_recursive(self):
        self.assertFalse(
            self.config.parts.get_reverse_dependencies(
                "nested-dependent", recursive=True
            )
        )
        self.assert_part_names(
            {"nested-dependent"},
            self.config.parts.get_reverse_dependencies("dependent", recursive=True),
        )
        self.assert_part_names(
            {"dependent", "nested-dependent"},
            self.config.parts.get_reverse_dependencies("main", recursive=True),
        )

    def test_dependency_loop(self):
        snapcraft_yaml = dedent(
            """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            grade: stable

            parts:
              p1:
                plugin: nil
                after: [p2]
              p2:
                plugin: nil
                after: [p1]
        """
        )
        raised = self.assertRaises(
            errors.SnapcraftLogicError, self.make_snapcraft_project, snapcraft_yaml
        )

        self.assertThat(
            raised.message,
            Equals("circular dependency chain found in parts definition"),
        )


class FilesetsTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.properties = {"filesets": {"1": ["1", "2", "3"], "2": []}}

    def test_expand_var(self):
        self.properties["stage"] = ["$1"]

        fs = _config._expand_filesets_for("stage", self.properties)
        self.assertThat(fs, Equals(["1", "2", "3"]))

    def test_no_expansion(self):
        self.properties["stage"] = ["1"]

        fs = _config._expand_filesets_for("stage", self.properties)
        self.assertThat(fs, Equals(["1"]))

    def test_invalid_expansion(self):
        self.properties["stage"] = ["$3"]

        raised = self.assertRaises(
            errors.SnapcraftLogicError,
            _config._expand_filesets_for,
            "stage",
            self.properties,
        )

        self.assertThat(
            str(raised),
            Contains(
                "'$3' referred to in the 'stage' fileset but it is not in filesets"
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import multiprocessing
import os
import platform
import sys
from typing import Set

from snapcraft import file_utils
from snapcraft.internal import common, errors, os_release

logger = logging.getLogger(__name__)


_ARCH_TRANSLATIONS = {
    "aarch64": {
        "kernel": "arm64",
        "deb": "arm64",
        "uts_machine": "aarch64",
        "cross-compiler-prefix": "aarch64-linux-gnu-",
        "cross-build-packages": ["gcc-aarch64-linux-gnu", "libc6-dev-arm64-cross"],
        "triplet": "aarch64-linux-gnu",
        "core-dynamic-linker": "lib/ld-linux-aarch64.so.1",
    },
    "armv7l": {
        "kernel": "arm",
        "deb": "armhf",
        "uts_machine": "arm",
        "cross-compiler-prefix": "arm-linux-gnueabihf-",
        "cross-build-packages": ["gcc-arm-linux-gnueabihf", "libc6-dev-armhf-cross"],
        "triplet": "arm-linux-gnueabihf",
        "core-dynamic-linker": "lib/ld-linux-armhf.so.3",
    },
    "i686": {
        "kernel": "x86",
        "deb": "i386",
        "uts_machine": "i686",
        "triplet": "i386-linux-gnu",
    },
    "ppc": {
        "kernel": "powerpc",
        "deb": "powerpc",
        "uts_machine": "powerpc",
        "cross-compiler-prefix": "powerpc-linux-gnu-",
        "cross-build-packages": ["gcc-powerpc-linux-gnu", "libc6-dev-powerpc-cross"],
        "triplet": "powerpc-linux-gnu",
    },
    "ppc64le": {
        "kernel": "powerpc",
        "deb": "ppc64el",
        "uts_machine": "ppc64el",
        "cross-compiler-prefix": "powerpc64le-linux-gnu-",
        "cross-build-packages": [
            "gcc-powerpc64le-linux-gnu",
            "libc6-dev-ppc64el-cross",
        ],
        "triplet": "powerpc64le-linux-gnu",
        "core-dynamic-linker": "lib64/ld64.so.2",
    },
    "riscv64": {
        "kernel": "riscv64",
        "deb": "riscv64",
        "uts_machine": "riscv64",
        "cross-compiler-prefix": "riscv64-linux-gnu-",
        "cross-build-packages": ["gcc-riscv64-linux-gnu", "libc6-dev-riscv64-cross"],
        "triplet": "riscv64-linux-gnu",
        "core-dynamic-linker": "lib/ld-linux-riscv64-lp64d.so.1",
    },
    "s390x": {
        "kernel": "s390",
        "deb": "s390x",
        "uts_machine": "s390x",
        "cross-compiler-prefix": "s390x-linux-gnu-",
        "cross-build-packages": ["gcc-s390x-linux-gnu", "libc6-dev-s390x-cross"],
        "triplet": "s390x-linux-gnu",
        "core-dynamic-linker": "lib/ld64.so.1",
    },
    "x86_64": {
        "kernel": "x86",
        "deb": "amd64",
        "uts_machine": "x86_64",
        "triplet": "x86_64-linux-gnu",
        "core-dynamic-linker": "lib64/ld-linux-x86-64.so.2",
    },
}


_32BIT_USERSPACE_ARCHITECTURE = {
    "aarch64": "armv7l",
    "armv8l": "armv7l",
    "ppc64le": "ppc",
    "x86_64": "i686",
}


_WINDOWS_TRANSLATIONS = {"AMD64": "x86_64"}


_HOST_CODENAME_FOR_BASE = {"core18": "bionic", "core": "xenial"}
_HOST_COMPATIBILITY = {
    "xenial": ["trusty", "xenial"],
    "bionic": ["trusty", "xenial", "bionic"],
}
_STATIC_BASES = ["bare"]

# TODO: just check the base.
_LINKER_VERSION_FOR_BASE = {"core20": "2.31", "core18": "2.27", "core": "2.23"}


def _get_platform_architecture():
    architecture = platform.machine()

    # Translate the windows architectures we know of to architectures
    # we can work with.
    if sys.platform == "win32":
        architecture = _WINDOWS_TRANSLATIONS.get(architecture)

    if platform.architecture()[0] == "32bit":
        userspace = _32BIT_USERSPACE_ARCHITECTURE.get(architecture)
        if userspace:
            architecture = userspace

    return architecture


class ProjectOptions:
    @property
    def parallel_build_count(self) -> int:
        try:
            build_count = len(os.sched_getaffinity(0))
        except AttributeError:
            # Fall back to multiprocessing.cpu_count()...
            try:
                build_count = multiprocessing.cpu_count()
            except NotImplementedError:
                logger.warning(
                    "Unable to determine CPU count; disabling parallel builds"
                )
                build_count = 1

        return build_count

    @property
    def is_cross_compiling(self):
        return self.__target_machine != self.__platform_arch

    @property
    def target_arch(self):
        return self.__target_arch

    @property
    def cross_compiler_prefix(self):
        try:
            # cross-compilation of x86 32bit binaries on a x86_64 host is
            # possible by reusing the native toolchain - let Kbuild figure
            # it out by itself and pass down an empty cross-compiler-prefix
            # to start the build
            if self.__platform_arch == "x86_64" and self.__target_machine == "i686":
                return ""
            return self.__machine_info["cross-compiler-prefix"]
        except KeyError:
            raise errors.SnapcraftEnvironmentError(
                "Cross compilation not supported for target arch {!r}".format(
                    self.__target_machine
                )
            )

    @property
    def additional_build_packages(self):
        packages = []
        if self.is_cross_compiling:
            packages.extend(self.__machine_info.get("cross-build-packages", []))
        return packages

    @property
    def arch_triplet(self):
        return self.__machine_info["triplet"]

    @property
    def deb_arch(self):
        return self.__machine_info["deb"]

    @property
    def kernel_arch(self):
        return self.__machine_info["kernel"]

    @property
    def parts_dir(self) -> str:
        return self._parts_dir

    @property
    def stage_dir(self) -> str:
        return self._stage_dir

    @property
    def prime_dir(self) -> str:
        return self._prime_dir

    @property
    def debug(self):
        return self._debug

    def __init__(
        self, target_deb_arch=None, debug=False, *, work_dir: str = None
    ) -> None:

        # Here for backwards compatibility.
        project_dir = os.getcwd()
        if work_dir is None:
            work_dir = project_dir

        self._debug = debug

        self._parts_dir = os.path.join(work_dir, "parts")
        self._stage_dir = os.path.join(work_dir, "stage")
        self._prime_dir = os.path.join(work_dir, "prime")

        logger.debug("Parts dir {}".format(self._parts_dir))
        logger.debug("Stage dir {}".format(self._stage_dir))
        logger.debug("Prime dir {}".format(self._prime_dir))

        self._set_machine(target_deb_arch)

    def _get_content_snaps(self) -> Set[str]:
        """Temporary shim for unit tests using ProjectOptions
        where Project is really required.  Will be removed in
        future convergence work.
        """
        return set()

    def _get_provider_content_dirs(self) -> Set[str]:
        """Temporary shim for unit tests using ProjectOptions
        where Project is really required.  Will be removed in
        future convergence work.
        """
        return set()

    def _get_stage_packages_target_arch(self) -> str:
        """Stub for 'Project' interface for tests using ProjectOptions()."""
        return self.deb_arch

    def is_static_base(self, base: str) -> bool:
        """Return True if a base that is intended to be static is used.

        Static bases require all their necessary components to live within
        the snap.
        """
        return base in _STATIC_BASES

    def is_host_compatible_with_base(self, base: str) -> bool:
        """Determines if the host is compatible with the GLIBC of the base.

        The system should warn early on when building using a host that does
        not match the intended base, this mechanism here enables additional
        logic when that is ignored to determine built projects will actually
        run.

        :param str base: the base core snap to search for linker.
        :returns: True if there are no GLIBC incompatibilities with the chosen
                  build host, else it returns False.
        :rtype: bool
        """
        try:
            codename = os_release.OsRelease().version_codename()
        except errors.OsReleaseCodenameError:
            return False

        logger.debug("Running on {!r}".format(codename))

        build_host_for_base = _HOST_CODENAME_FOR_BASE.get(base)
        if build_host_for_base is None:
            return False

        compatible_hosts = _HOST_COMPATIBILITY.get(build_host_for_base, [])

        return codename in compatible_hosts

    # This is private to not make the API public given that base
    # will be part of the new Project.
    def _get_linker_version_for_base(self, base: str) -> str:
        """Returns the linker version for base."""
        try:
            return _LINKER_VERSION_FOR_BASE[base]
        except KeyError:
            linker_file = os.path.basename(self.get_core_dynamic_linker(base))
            return file_utils.get_linker_version_from_file(linker_file)

    def get_core_dynamic_linker(self, base: str, expand: bool = True) -> str:
        """Returns the dynamic linker used for the targeted core.

        :param str base: the base core snap to search for linker.
        :param bool expand: expand the linker to the actual linker if True,
                            else the main entry point to the linker for the
                            projects architecture.
        :return: the absolute path to the linker
        :rtype: str
        :raises snapcraft.internal.errors.SnapcraftMissingLinkerInBaseError:
            if the linker cannot be found in the base.
        :raises snapcraft.internal.errors.SnapcraftEnvironmentError:
            if a loop is found while resolving the real path to the linker.
        """
        core_path = common.get_installed_snap_path(base)
        dynamic_linker_path = os.path.join(
            core_path,
            self.__machine_info.get("core-dynamic-linker", "lib/ld-linux.so.2"),
        )

        # return immediately if we do not need to expand
        if not expand:
            return dynamic_linker_path

        # We can't use os.path.realpath because any absolute symlinks
        # have to be interpreted relative to core_path, not the real
        # root.
        seen_paths = set()  # type: Set[str]
        while True:
            if dynamic_linker_path in seen_paths:
                raise errors.SnapcraftEnvironmentError(
                    "found symlink loop resolving dynamic linker path"
                )

            seen_paths.add(dynamic_linker_path)
            if not os.path.lexists(dynamic_linker_path):
                raise errors.SnapcraftMissingLinkerInBaseError(
                    base=base, linker_path=dynamic_linker_path
                )
            if not os.path.islink(dynamic_linker_path):
                return dynamic_linker_path

            link_contents = os.readlink(dynamic_linker_path)
            if os.path.isabs(link_contents):
                dynamic_linker_path = os.path.join(core_path, link_contents.lstrip("/"))
            else:
                dynamic_linker_path = os.path.join(
                    os.path.dirname(dynamic_linker_path), link_contents
                )

    def _set_machine(self, target_deb_arch):
        self.__platform_arch = _get_platform_architecture()
        if not target_deb_arch:
            self.__target_machine = self.__platform_arch
        else:
            self.__target_machine = _find_machine(target_deb_arch)
            logger.info("Setting target machine to {!r}".format(target_deb_arch))

        self.__machine_info = _ARCH_TRANSLATIONS[self.__target_machine]

        # Set target arch to match the host if unspecified.
        if target_deb_arch is None:
            self.__target_arch = self.__machine_info.get("deb")
        else:
            self.__target_arch = target_deb_arch


def _get_deb_arch(machine):
    return _ARCH_TRANSLATIONS[machine].get("deb", None)


def _find_machine(deb_arch):
    for machine in _ARCH_TRANSLATIONS:
        if _ARCH_TRANSLATIONS[machine].get("deb", "") == deb_arch:
            return machine
        elif _ARCH_TRANSLATIONS[machine].get("uts_machine", "") == deb_arch:
            return machine

    raise errors.SnapcraftEnvironmentError(
        "Cannot set machine from deb_arch {!r}".format(deb_arch)
    )
